#
# Copyright (c) 2010-2011, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
#
#
# Author: Nick Blundell <blundeln [AT] gmail [DOT] com>
# Organisation: www.nickblundell.org.uk
# 
# Description:
#   Defines a Rollbackable class and util functions.
#

import copy
from debug import *
from exceptions import *


class Rollbackable(object) :
  """
  A class that can have its state rolled back, to undo modifications.
  A blanket deepcopy is not ideal, though we can explore more efficient
  solutions later (e.g. copy-before-modify).
  """

  # XXX: Do we always need to copy on get AND set? Have to careful that original state is not set.
  # XXX: Basically need to make sure that original state cannot be modified
  # XXX: Perhaps add copy-flag
  def _get_state(self, copy_state=True) :
    """
    Gets the state of this object that is required for rollback.  This is a
    catch-all function for classes that specialise Rollbackable, though where
    possible more efficient functions should be implemented.

    Usually we will wish to obtain a copy, so the original state is not
    modified, though sometimes (e.g. when comparing states) we will not require
    a copy.
    """
    if copy_state :
      return copy.deepcopy(self.__dict__)
    else :
      return self.__dict__
      

  def _set_state(self, state, copy_state=True) :
    """
    Sets the state of this object for rollback.  This is a
    catch-all function for classes that specialise Rollbackable, though where
    possible more efficient functions should be implemented.

    Usually we will wish to set a copy, so the original state is not
    modified, though sometimes we will not require
    a copy (e.g. if we know the original state will no longer be required).
    """
    if copy_state :
      self.__dict__ = copy.deepcopy(state)
    else :
      self.__dict__ = state


  def __eq__(self, other):
    """So we can easily compare if two objects have state of equal value."""
    # TODO: To use this is expensive and should be replaced by a more
    # efficient method
    # TODO:   perhaps a dirty-flag scheme???
    return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

  @staticmethod
  def TESTS():

    class SomeClass(Rollbackable):
      def __init__(self, x, y) :
        self.x, self.y = x, y

    o = SomeClass(1, [3,4])
    state1 = o._get_state()
    o.x = 3
    o.y.append(16)
    assert(o.x == 3)
    assert(o.y == [3,4,16])

    o._set_state(state1)
    assert(o.x == 1)
    assert(o.y == [3,4])

    # Test value comparision.
    o1 = SomeClass(1, [3,4])
    o2 = SomeClass(1, [3,4])
    assert(o1 == o2)
    o2.y[1] = 9
    assert(o1 != o2)
    o2.y[1] = 4
    assert(o1 == o2)



#
# Utility functions for getting and setting the state of multiple rollbackables.
#

def get_rollbackables_state(*rollbackables, **kargs) :
  """Handy function to get the state of multiple rollbackables, conviently ignoring those with value None."""
  # Assume we copy state, unless directed otherwise.
  if "copy_state" in kargs and kargs["copy_state"] == False:
    copy_state = False
  else :
    copy_state = True

  # Note: rollbackables must be in same order for get and set.
  rollbackables_state = []
  for rollbackable in rollbackables :
    if isinstance(rollbackable, Rollbackable) :
      rollbackables_state.append(rollbackable._get_state(copy_state=copy_state))
  
  #if IN_DEBUG_MODE :
  #  d("Getting state : %s" % rollbackables_state)
  
  return rollbackables_state

def set_rollbackables_state(new_rollbackables_state, *rollbackables, **kargs) :
  """Handy function to set the state of multiple rollbackables, conviently ignoring those with value None."""
  # Assume we copy state, unless directed otherwise.
  if "copy_state" in kargs and kargs["copy_state"] == False:
    copy_state = False
  else :
    copy_state = True

  #if IN_DEBUG_MODE :
  #  d("Setting state to: %s" % new_rollbackables_state)

  state_index = 0
  for rollbackable in rollbackables:
    if isinstance(rollbackable, Rollbackable) :
      rollbackable._set_state(new_rollbackables_state[state_index], copy_state=copy_state)
      state_index += 1


class automatic_rollback:
  """
  Allows rollback of reader state using the 'with' statement, for cleaner
  syntax.

  Possible extensions:
  """
  
  def __init__(self, *rollbackables, **kargs) :
    # Store the rollbackables. Note, for convenience, allow rollbackables to be None (i.e. store only Reader instances)
    self.some_state_changed = False
    self.check_for_state_change = "check_for_state_change" in kargs and kargs["check_for_state_change"] or None
    # Allows initial state to be reused.
    self.initial_state = "initial_state" in kargs and kargs["initial_state"] or None
    self.rollbackables = rollbackables
  
  def __enter__(self) :
    # Store the start state of each reader, unless we have been passed some
    # initial state to reuse.
    if self.initial_state :
      self.start_state = self.initial_state
    else :
      self.start_state = get_rollbackables_state(*self.rollbackables)
  
  def __exit__(self, type, value, traceback) :
    # If a RollbackException is thrown, revert all of the rollbackables.
    if type and issubclass(type, RollbackException) :
      set_rollbackables_state(self.start_state, *self.rollbackables)
      d("Rolled back rollbackables to: %s." % str(self.rollbackables))
   
    # XXX: Optimise this to first check for concrete reader.
    if self.check_for_state_change :
      # Not changing this state, so no need to copy it.
      current_state = get_rollbackables_state(*self.rollbackables, copy_state=False)
      #d("State: start: %s current: %s" % (self.start_state, current_state))
      self.some_state_changed = current_state != self.start_state

    # Note, by not returning True, we do not supress the exception, which gives
    # us maximum flexibility.


  @staticmethod
  def TESTS() :
    class SomeClass(Rollbackable):
      def __init__(self, x, y) :
        self.x, self.y = x, y

    o_1 = SomeClass(1, [3,4])
    o_2 = None                # Important that we can handle None to simplify code.
    o_3 = SomeClass(1, [3,4])
   
    try :
      with automatic_rollback(o_1, o_2, o_3):
        o_1.x = 3
        o_3.y.append(16)
        assert(o_1.x == 3)
        assert(o_3.y == [3,4,16])
        raise LensException() # In practice we will usually use LensException
    except LensException:
      pass # Don't wish to stop test run.
       
    # Check we rolled back.
    assert(o_1.x == 1)
    assert(o_3.y == [3,4])





#
# Copyright (c) 2010, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
#
#
# Author: Nick Blundell <blundeln [AT] gmail [DOT] com>
# Organisation: www.nickblundell.org.uk
# 
# Description:
# 
#
from debug import *

# Thrown when tentative object state should be rolled back.
class RollbackException(Exception): pass


class LensException(RollbackException):
  """
  Thrown when parsing or creating lenses to trigger rollback, such that parsing
  may resume at a higher level (e.g. to try another lens path), if possible.
  """

  def __init__(self, msg=None):
    self.__msg = msg
    if IN_DEBUG_MODE :
      d("Throwing: %s (from %s)" % (self.__msg, self.get_thrown_from()))

  def get_thrown_from(self) :
    
    import inspect
    from nbdebug import getCallerLocation
    
    #TODO: Could tidy this up and perhaps integrate with nbdebug.

    ignore_frames = ["lens_assert()", "LensException.get_thrown_from()", "LensException.__init__()"]
    callerFrame = inspect.currentframe()
    location = None
    while callerFrame:
      location = getCallerLocation(callerFrame)
      if location not in ignore_frames :
        break
      callerFrame = callerFrame.f_back
    
    return location

  def __str__(self):
      return "LensException: %s" % self.__msg

# Thrown when an abstract token collection cannot find an appropriate token in the
# PUT direction.
# Note, when deciding whether to throw a LensException or Exception it is useful
# to consider the Or lens, when alternate branches may be tried (e.g. is it a
# problem with the lens definition or just a failed parsing branch)
class NoTokenToConsumeException(LensException): pass
class NoDefaultException(LensException): pass
class TooFewIterationsException(LensException): pass
class NotFullyConsumedException(LensException): pass

# Thrown when it looks like a lens may iterate infinitely.
class InfiniteIterationException(LensException) : pass   #XXX Deprecated

class InfiniteRecursionException(Exception): pass
class CannotStoreException(Exception): pass

class EndOfStringException(LensException):
  pass



#
# Copyright (c) 2010-2011, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# Author: Nick Blundell <blundeln [AT] gmail [DOT] com>
# Organisation: www.nickblundell.org.uk
# 
#
import inspect
import sys
from exceptions import *
from containers import *
from readers import *
from util import *
from debug import *
from base_lenses import *


class Forward(Lens):
  """
  Allows forward declaration of a lens, which may be bound later, primarily to
  allow for lens recursion.  Based on the idea used in pyparsing, since we must
  define variables before we use them, unless we use some python interpreter
  pre-processing.
  """
  def __init__(self, recursion_limit=100, **options):
    super(Forward, self).__init__(**options)
    d("Creating")
    self.recursion_limit = recursion_limit
  
  def bind_lens(self, lens) :
    d("Binding to lens %s" % lens)
    assert_msg(len(self.lenses) == 0, "The lens cannot be re-bound.")
    self.set_sublens(lens)
  
  def _get(self, *args, **kargs) :
    assert_msg(len(self.lenses) == 1, "A lens has yet to be bound.")
    return self.lenses[0]._get(*args, **kargs)

  def _put(self, *args, **kargs) :
    assert_msg(len(self.lenses) == 1, "A lens has yet to be bound.")
    
    # Ensure the recursion limit is set before we start this.
    original_limit = sys.getrecursionlimit()
    if self.recursion_limit :
      sys.setrecursionlimit(self.recursion_limit)
    
    try :
      output = self.lenses[0]._put(*args, **kargs)
    except RuntimeError:
      raise InfiniteRecursionException("You will need to alter your grammar, perhaps changing the order of Or lens operands")
    finally :
      sys.setrecursionlimit(original_limit)
    
    return output


  # Use the lshift operator, as does pyparsing, since we cannot easily override (re-)assignment.
  def __lshift__(self, other) :
    assert_msg(isinstance(other, Lens), "Can bind only to a lens.")
    self.bind_lens(other)


  @staticmethod
  def TESTS():

    # TODO: Need to think about semantics of this and how it will work with groups.
    d("GET")
    lens = Forward()
    # Now define the lens (must use '<<' rather than '=', since cannot easily
    # override '=').
    lens << "[" + (AnyOf(alphas, type=str) | lens) + "]"
    
    # Ensure the lens is enclosed in a container lens.
    lens = Group(lens, type=list)
    
    got = lens.get("[[[h]]]")
    assert(got == ["h"])
    
    d("PUT")
    got[0] = "p"
    output = lens.put(got)
    assert(output == "[[[p]]]")

    # Note that this lens results in infinite recursion upon CREATE.
    d("CREATE")
    output = lens.put(["k"])
    assert(output == "[k]")
    
    # If we alter the grammar slightly, we will get an infinite recursion error,
    # since the lens could recurse to an infinite depth before considering the
    # AnyOf() lens.
    lens = Forward() 
    lens << "[" + (lens | AnyOf(alphas, type=str)) + "]"
    lens = Group(lens, type=list)
    with assert_raises(InfiniteRecursionException) :
      output = lens.put(["k"])
    

class Until(Lens) :
  """
  Match anything up until the specified lens.  This is useful for lazy parsing,
  but not the be overused (e.g. chaining can be bad: Until("X") + Until("Y")!).
  
  # TODO: Could parse lens but not include in stored string.
  """
  
  def __init__(self, lens, include_lens=False, **options):
    """
    Arguments:
      include_lens - Set to true if the specified lens should also be consumed.
    """
    super(Until, self).__init__(**options)
    self.set_sublens(lens)
    self.include_lens = include_lens

  def _get(self, concrete_input_reader, current_container, force_return=False) :
    # Note, we add force_return here so that put can utilise output regardless of
    # whether or not this is a STORE lens.  Really I should create a parsing
    # function which both get and put use.
    
    # Remember the input position before we start to consume chars.
    initial_position = concrete_input_reader.get_pos()
    
    stopping_lens = self.lenses[0]
   
    while True :
      start_state = get_rollbackables_state(concrete_input_reader)
      try :
        stopping_lens.get(concrete_input_reader)

        # If we are not to include consumption of the lenes, roll back the state
        # after successfully getting the lens, since we do not want to include
        # consumption of the lens.
        if not self.include_lens :
          d("Rollbacked from %s" % get_rollbackables_state(concrete_input_reader))
          set_rollbackables_state(start_state, concrete_input_reader)
          d("Rollbacked to %s" % get_rollbackables_state(concrete_input_reader))
        else :
          pass

        break
      
      except LensException :
        # We have not reached the stopping lens in input yet, so we rollback and then carry on.
        d("stopping_lens failed soi continuing.")
        set_rollbackables_state(start_state, concrete_input_reader)
      
      # Advance the input reader by one char - this will form part of our lens' GOTen string.
      try :
        concrete_input_reader.consume_char()
      except EndOfStringException:
        # Break if we reach the end of the input.
        break

    parsed_chars = concrete_input_reader.get_consumed_string(initial_position) 

    if not parsed_chars :
      raise LensException("Expected to get at least one character!")
  
    if self.has_type() or force_return :
      return parsed_chars
    
    # Return nothing if we are not a STORE lens.
    return None


  def _put(self, item, concrete_input_reader, current_container) :
    if self.has_type() :
      if not isinstance(item, str) and len(item) > 0:
        raise Exception("Expected to be passed a string of length at least one character, not %s." % item)
      
      # Consume input.
      if concrete_input_reader :
        self.get(concrete_input_reader)
      output = item
    else :
      if has_value(item):
        raise LensException("As a non-STORE lens, %s did not expect to be passed an item %s to PUT." % (self, item))
      
      # Use output from input, or fail if we have no concrete input.
      if concrete_input_reader :
        output = self._get(concrete_input_reader, None, force_return=True)
      else :
        raise NoDefaultException("Cannot CREATE: a default should have been set on lens %s, or an outer lens." % self)

    return output


  @staticmethod
  def TESTS() :
    d("GET")
    lens = Group("("+Until(")", type=str) + ")", type=list)
    got = lens.get("(in the middle)")
    assert(got == ["in the middle"])

    d("PUT")
    output = lens.put(["monkey"])
    assert(output == "(monkey)")
   
    test_description("Try with include_lens=True")
    assert(lens.get(lens.put(["monkey"])) == ["monkey"])
    lens = Group("("+Until(")", type=str, include_lens=True), type=list)
    got = lens.get("(in the middle)")
    assert(got == ["in the middle)"])

    # XXX: Perhaps protect against this, or perhaps leave to lens user to worry about?!
    #assert(lens.get(lens.put(["mon)key"])) == ["monkey"])



#
# Copyright (c) 2010-2011, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
#
#
# Author: Nick Blundell <blundeln [AT] gmail [DOT] com>
# Organisation: www.nickblundell.org.uk
# 
# Description:
#   Isolates all debugging functions.
#

# Optionally include nbdebug functions.
try :
  from nbdebug import d, breakpoint, set_indent_function, IN_DEBUG_MODE
except :
  d = lambda x: None
  set_indent_function = None
  IN_DEBUG_MODE = False

from exceptions import *

# More syntacticaly consistant assert function, for displaying explanations
def assert_msg(condition, msg=None) :
  assert condition, (msg or "")

def test_description(msg) :
  """A debug message that will stand out."""
  msg = "========= " + msg + " ========="
  return d(msg)

def assert_equal(got, expected) :
  assert_msg(got == expected, "Expected >>>%s<<< but got >>>%s<<<" % (expected, got))

class assert_raises:
  """A cleaner way to assert that an exception is thrown from some code."""
  
  def __init__(self, exception_class) :
    self.exception_class = exception_class
  
  def __enter__(self) :
    pass
  
  def __exit__(self, type, exception, traceback) :
    if not exception :
      raise Exception("Expected to see exception: %s" % self.exception_class)
    
    # Returning True means 'suppress exception', which we do if the exception is
    # of the type we expected.
    return isinstance(exception, self.exception_class)

  @staticmethod
  def TESTS() :
    d("Testing")
    
    # Assert the ZeroDivisionError is thrown.
    with assert_raises(ZeroDivisionError) :
      x = 1 / 0

    # Assert that we expected the ZeroDivisionError to be thrown.
    with assert_raises(Exception) :
      with assert_raises(ZeroDivisionError) :
        x = 1 / 1
    
    # Confirm that the unexpected exception is let through.  My most beautiful test, ever!
    with assert_raises(IndexError) :
      with assert_raises(ZeroDivisionError) :
        x = []
        x[0] = 2

def auto_name_lenses(local_variables) :
  """
  Gives names to lenses based on their local variable names, which is
  useful for tracing parsing. Should be called with globals()/locals()
  """
  from pylens.base_lenses import Lens
  for variable_name, obj in local_variables.iteritems() :
    if isinstance(obj, Lens) :
      obj.name = variable_name


# Set an debug message indentation function if the debug library is in use.
if set_indent_function :
  
  def debug_indent_function() :
    """
    Nicely indents the debug messages according to the hierarchy of lenses.
    """ 
    import inspect
    # Create a list of all function names in the trace.
    function_names = []

    # Prepend the callers location to the message.
    callerFrame = inspect.currentframe()
    while callerFrame :
      location = callerFrame.f_code.co_name
      function_names.append(location)
      callerFrame = callerFrame.f_back

    indent = 0
    # Includes 'get' and 'put' since get may be called directly in put (not _put), etc.
    for name in ["_put", "_get", "put", "get"] :
      indent += function_names.count(name)
    indent -= 1
    indent = max(0, indent)
   
    return " "*indent

  set_indent_function(debug_indent_function)




#
# Copyright (c) 2010, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# Author: Nick Blundell <blundeln [AT] gmail [DOT] com>
# Organisation: www.nickblundell.org.uk
# 
# Description:
#   Stateful string reader classes (i.e. that can be rolled back for tentative parsing)
#

from debug import *
from exceptions import *
from util import *
from containers import *


class ConcreteInputReader(Rollbackable):
  """Stateful reader of the concrete input string."""

  def __init__(self, input_string):
    
    # If input_string is in fact a ConcreteInputReader, copy its state.
    if isinstance(input_string, self.__class__) :
      self.position = input_string.position
      self.string = input_string.string
    # Otherwise, initialise our state.
    else :
      assert(isinstance(input_string, str))
      self.position  = 0
      self.string    = input_string

  def reset(self) :
    self.set_pos(0)

  def _get_state(self, copy_state=True) :
    return self.get_pos()

  def _set_state(self, state, copy_state=True) :
    self.set_pos(state)

  def get_consumed_string(self, start_pos=0) :
    return self.string[start_pos:self.position]

  def get_pos(self):
    return self.position

  def set_pos(self, pos) :
    assert(isinstance(pos, int))
    self.position = pos

  def get_remaining(self) :
    """Return the text that remains to be parsed - useful for debugging."""
    return self.string[self.position:]

  def consume_string(self,length):
    """
    Consume a string of specified length from the input.
    """
    if self.position+length > len(self.string):
      raise EndOfStringException()
    start = self.position
    self.position += length
    return self.string[start:self.position]


  def consume_char(self):
    """
    Consume and return the next char from input.
    """
    if self.is_fully_consumed():
      raise EndOfStringException()
    char = self.string[self.position]
    self.position += 1
    return char


  def is_fully_consumed(self):
    """
    Return whether the string is fully consumed
    """
    return self.position >= len(self.string)


  def is_aligned_with(self, other) :
    """Check if this reader is aligned with another."""
    return self.position == other.position and self.string == other.string

  def __str__(self) :
    # Return a string representation of this reader, to help debugging.
    if self.is_fully_consumed() :
      return "END_OF_STRING"

    display_string = self.string[self.position:]
    return "'" + truncate(display_string) + "'"
  __repr__ = __str__

  @staticmethod
  def TESTS() :
    concrete_reader = ConcreteInputReader("ABCD")
    output = ""
    for i in range(0,2) :
      output += concrete_reader.consume_char()
    assert not concrete_reader.is_fully_consumed()
    assert concrete_reader.get_remaining() == "CD"
    assert concrete_reader.get_consumed_string(0) == "AB"
    
    for i in range(0,2) :
      output += concrete_reader.consume_char()
    assert output == "ABCD"
    assert concrete_reader.is_fully_consumed()
    
    concrete_reader = ConcreteInputReader("ABCD")

    # Now test with rollback.
    concrete_reader = ConcreteInputReader("ABCD")
    try :
      with automatic_rollback(concrete_reader):
        concrete_reader.consume_char()
        assert concrete_reader.get_remaining() == "BCD"
        raise LensException()
    except LensException:
      pass # Don't want to stop tests.

    assert concrete_reader.get_remaining() == "ABCD"

    # Test that clones share the string object, for efficiency.
    cloned_reader = ConcreteInputReader(concrete_reader)
    assert(cloned_reader.string is concrete_reader.string)
    assert(cloned_reader.is_aligned_with(concrete_reader))
    cloned_reader.position += 1
    assert(not cloned_reader.is_aligned_with(concrete_reader))





#
# Copyright (c) 2010-2011, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# Author: Nick Blundell <blundeln [AT] gmail [DOT] com>
# Organisation: www.nickblundell.org.uk
# 
#

"""
These lenses form the core of pylens.  The Lens class is the base of all other
lenses and encapsulates (and hides) much of the complexity of the framework
whilst allowed us to extend it.
"""

import inspect

from debug import *
from settings import *
from exceptions import *
from containers import *
from readers import *
from item import *
from util import *
from charsets import *


#########################################################
# Base Lens
#########################################################

class Lens(object) :
  """Base lens, which all other lenses extend."""
  
  def __init__(self, type=None, name=None, default=None, **options) :
    """
    Arguments:
      
      - type: The native type of this lens (e.g. int, str, some LensObject
        class).  If set, this is a STORE lens, otherwise a NON-STORE lens.
      - name: For assigning a name to this lens for debugging purposes.
      - default: the default output of this lens if a NON-STORE lens.
      - options: arbitrary keyword options can be set on a lens to affect its
        functionality in an extendible way.  For example, we might use this to
        specify how we align data items to their original position in the
        source string, perhaps by some label, their original position, or by
        some order in the native type.
    """
    self.type = type

    # Set the default value of this lens (in the PUT direction) if it is not a
    # STORE lens.
    if has_value(type) and has_value(default) :
      raise Exception("Cannot set a default value on a lens with a type (i.e. on a store lens).")
    self.default = default
    
    # For debugging purposes, allow a friendly name to be given to the lens,
    # otherwise an automated name will be displayed (e.g. "And(106)")
    self.name = name

    # Composite lenses will store their sub-lenses in here, for consistency,
    # and later perhaps to allow for some reasonning about a lens' structure.
    self.lenses = []

    # Allow arbitrary arguments to be set on the lens which can aid flexible
    # storage and retrival of items from a container.
    self.options = Properties(**options)

    #
    # Argument shortcuts
    #

    # There is no point setting a non-store lens as a label or to have a
    # label, so assume the user wanted a lens type of str.
    if not has_value(self.type) and (self.options.is_label or self.options.label) :
      self.type = str



  def get(self, concrete_input, current_container=None) :
    """
    Returns a data item from the string *concrete_input* according to this
    lens and its type; if this lens has no type, the input string is simply
    consumed and nothing returned to be used in our abstract structure.

    This effectively wraps the _get function (GET proper) of the specific
    lens, handling all of the common tasks (e.g. input normalisation, creation
    of stateful item containers, rolling back state for failed parsing
    branches, etc.).
    
    Arguments:

      - concrete_input: Concrete input string to parse from.
      - current_container: Outer container (e.g. a list, dict, some LensObject
        class, etc.) into which items are being extracted.  Usually this is
        used only internally, as the parsing descends into sub-lenses, some of which
        define their own containers (e.g. using lenses such as Group) and some which
        don't but instead store items into some higher-level container.
    """
    #
    # Algorithm
    #
    # If we have a container type
    #   replace container with new one - to save hassle of doing it in each
    #   _get (proper), since a typed-container lens will always deal with its
    #   own container
    # item = _get(input, container)
    # if typed_lens:
    #   cast item if not instance
    #   assert(item is correct type)
    #   assert not already got properties
    #   set properties
    #
    # Note, even if non-typed, we may return an item from a lens we wrap (e.g. Or)
    #
    # item = process_item(item) # e.g. allow pre-processing of item (e.g. for auto_list)
    #
    # return item

    # Ensure we have the concrete input in the form of a ConcreteInputReader
    assert_msg(has_value(concrete_input), "Cannot GET if there is no input string!")
    concrete_input_reader = self._normalise_concrete_input(concrete_input)
  
    if IN_DEBUG_MODE :
      d("Initial state: in={concrete_input_reader}, cont={current_container}".format(
        concrete_input_reader = concrete_input_reader,
        current_container = current_container,
      ))

    # Remember the start position of the concrete reader, to aid
    # re-alignment of concrete structures when we Lens.put is later called.
    # We will store this in a returned items meta_data, effictively giving it
    # a lifeline back to where it came from.
    concrete_start_position = concrete_input_reader.get_pos()

    # Create an empty appropriate container class for our lens, if there is one;
    # this will be None if we are not a container-type lens (e.g. a dict or
    # list).
    lens_container = self._create_lens_container()
    if lens_container :
      d("Created container")

    # If we are a container-type lens, replace the current container of
    # sub-lenses with our own container (lens_container).
    if lens_container :
      # Call GET proper with our container, checking that no item is returned,
      # since all items should be stored WITHIN the container
      assert_msg(self._get(concrete_input_reader, lens_container) == None, "Container lens %s has GOT an item, but all items must be stored in the current container, not returned." % self)
      
      # Since we created the container, we will return it as our item, for a
      # higher order lens to store.
      item = lens_container.unwrap()
     


    # Otherwise, call GET proper using the outer container, if there is one.
    else :
      item = self._get(concrete_input_reader, current_container)

    # If we are a STORE lens (i.e. we extract an item) ...
    if self.has_type() :
      
      # Cast the item to our type (usually if it is a string being cast to a
      # simple type, such as int).
      assert_msg(has_value(item), "Somethings gone wrong: %s is a STORE lens, so we should have got an item." % self)
      if not isinstance(item, self.type) :
        item = self.type(item)

      # Double check we got an item of the correct type (after any casting).
      assert(isinstance(item, self.type))
     
      # Allow meta data to be stored on the item.
      item = enable_meta_data(item)
      
      # Now add source info to the item's meta data, which will help when
      # putting it back into a concrete structure.
      
      # A reference to the lens that extracted the item.
      item._meta_data.lens = self
      d("Set meta on %s to %s" % (item, item._meta_data))

      # A reference to the concrete reader and position parsed from.
      item._meta_data.concrete_start_position = concrete_start_position
      item._meta_data.concrete_end_position = concrete_input_reader.get_pos()
      item._meta_data.concrete_input_reader = concrete_input_reader

      # If the item was unwrapped from a container, update meta with label
      # from the container, which may have been set if there was an is_label
      # lens.
      if lens_container :
        item._meta_data.label = lens_container.get_label()

    # Note that, even if we are not a typed lens, we may return an item
    # extracted from some sub-lens, the Or lens being a good example.

    if IN_DEBUG_MODE :
      if has_value(item) :
        d("GOT: %s %s" % (item, item._meta_data.label and "[label: '%s']" % (item._meta_data.label) or ""))
      else :
        d("GOT: NOTHING (to store)")
  
    # If appropriate, check the input was fully consumed by this lens
    if isinstance(concrete_input, str) and GlobalSettings.check_consumption and not concrete_input_reader.is_fully_consumed() :
      raise NotFullyConsumedException("The following input remains to be consumed by this lens: '%s'" % concrete_input_reader.get_remaining())

    # Pre-process outgoing item.
    item = self._process_outgoing_item(item)

    return item


  def put(self, item=None, concrete_input=None, current_container=None, label=None) :
    """
    Puts an item from our abstract model through the lens to generate its
    string representation, weaving the model items back into an original
    string if there was one (i.e. if we are PUTTING rather than CREATING).

    Frankly, this is the big beast that drives pylens and you should not worry
    too much about understanding the internals if you simply wish to build lenses.

    Arguments:

      - item: - item to PUT.
      - concrete_input: concrete input for weaving between STORE lens values.
      - current_container - outer container from which items will be consumed
        and PUT back.
      - label - allows the user to set a label on the passed item, to allow for
        structures that internally contain a label.

    This effectively wraps the _put function (PUT proper) of the specific
    lens, handling all of the common tasks (e.g. input normalisation, creation
    of stateful containers, rolling back state for failed parsing branches).

    Note that we make no distinction between PUT and CREATE (from the
    literatur): since a previously extracted item will carry information of
    its concrete structure within its meta data it will use this for weaving
    in non-stored artifacts; otherwise, default artifacts will be used (as in
    CREATE).

    In some cases the order that items are put back in will differ from their
    original order, so we are careful to both consume and discard from the
    outer concrete structure whilst PUTTING with the item' own concrete
    structure.  If both structures are aligned (i.e. the item goes back in its
    original order) then a single PUT on the outer concrete reader is
    performed.
    """
   
    #
    # Algorithm
    #
    # Assert item => container == None
    # Note, though an item will hold its own input, the lens must consumed from
    # the outer input reader if one is supplied.
    #
    # Normalise concrete input reader.
    #
    # If we are an un-typed lens
    #   if we have a default value and no concrete input, return it.
    #   otherwise return put proper, passing through our args - note, item could be passed through us
    #
    # Now assume we are typed lens
    #
    #  If we are passed an item
    #    ensure meta enabled (e.g. for a new item, not previously extracted)
    #    pre-process item (e.g. to handle auto_list)
    #    check correct type, else raise LensException
    #    get the item input reader if there is one - can be None
    #    if item_input_reader 
    #      if input_reader
    #        if item_input_reader is not aligned with (outer) input_reader
    #          consume from input_reader
    #          set input_reader item_input_reader
    #        else we use the input_reader (i.e both consume and put) - can discard item_input_reader
    #    else :
    #      No input for this item, so we are CREATING
    #      if input_reader
    #        consume from input_reader
    #      set input_reader = None
    #      
    #    if we are container type, wrap item as current_container and set item = None
    #    call put proper on item with item, input_reader and current_container
    #  else if we are passed a container (and no item)
    #    instruct the container to put it.
    #
    #  Should have returned by here, so raise LensException: expected something to put.
    #

    # If we are passed an item, we do not expect an outer container to also have
    # been passed.
    if has_value(item) :
      assert_msg(current_container == None, "A lens should not be passed both a container and an item.")
      # Note, however, that a typed item may well be passed an outer concrete
      # reader, which will be used for weaving in non-stored artifacts of concrete
      # structures.

    # Ensure we have a ConcreteInputReader; otherwise None.
    concrete_input_reader = self._normalise_concrete_input(concrete_input)
    
    # We need this for checking consumption, since concrete_input_reader can be
    # changed by our algorithm.
    original_concrete_input_reader = concrete_input_reader

    # Display some useful info for debug tracing.
    if IN_DEBUG_MODE :
      
      # It's very useful to see if an item holds a label in its meta.
      if hasattr(item, "_meta_data") and has_value(item._meta_data.label) :
        item_label_string = " [label: %s]" % item._meta_data.label
      else :
        item_label_string = ""
      
      d("Initial state: item={item}{item_label_string}, in={concrete_input_reader}, cont={current_container}".format(
        item = item,
        concrete_input_reader = concrete_input_reader,
        current_container = current_container,
        item_label_string = item_label_string,
      ))


    # First handle cases where our lens does not directly store an item: it will
    # either return the default output string, if it has one; or it will
    # generate some output internally, perhaps from the input or from the
    # default output of a sub-lens.
    if not self.has_type() :
      # Use default (for CREATE)
      if concrete_input_reader == None and has_value(self.default) :
        output = str(self.default)
      
      # Otherwise do a PUT proper, passing through our arguments, for example
      # our child lens may put an item directly or from the container of use
      # its own default value.
      else :
        output = self._put(item, concrete_input_reader, current_container)


    # Now we can assume that our lens has a type (i.e. will directly PUT an
    # item)
    elif has_value(item) :
    
      # For the sake of algorithmic consistancy, ensure the incoming item can
      # hold meta data.
      item = enable_meta_data(item)
     
      # Store original state of item, including meta data, so we can recover it on success,
      # since PUT can be destructive to some containers, depending on how they
      # are implemented and their meta data.
      original_meta_data = item._meta_data.copy()
      original_item = item
      if isinstance(item, Rollbackable) :
        original_state = item._get_state()

      # Associate a label with the item, usually a label passed from the user,
      # which is required internally by a structure.
      if has_value(label) :
        item._meta_data.label = label

      # Pre-process the incoming item (e.g to handle auto_list or other future
      # extensions)
      item = self._process_incoming_item(item)
     
      # Check our item's type is compatible with the lens.
      if not isinstance(item, self.type) :
        raise LensException("This lens %s of type %s cannot PUT an item of that type %s" % (self, self.type, type(item)))
      
      # If this item was previously GOTten, we can get its original input.
      if has_value(item._meta_data.concrete_input_reader) :
        
        # Create a personal concrete reader for this item, based on its meta
        # data.
        item_input_reader = ConcreteInputReader(item._meta_data.concrete_input_reader)
        item_input_reader.set_pos(item._meta_data.concrete_start_position)

        # If the readers are not aligned...
        if not(has_value(concrete_input_reader) and item_input_reader.is_aligned_with(concrete_input_reader)) :
          # Consumed from the outer reader, if there is one.
          if has_value(concrete_input_reader) :
            d("Inputs not aligned, so consuming and discarding from outer input reader.")
            self.get_and_discard(concrete_input_reader, current_container)

          # Now use substitute the outer reader (if there was one) with our
          # item's reader
          concrete_input_reader = item_input_reader
      
      else :
        
        # Otherwise, if our item had no source meta, we will be CREATING, but
        # must still consume from the outer reader, if there is one.
        if has_value(concrete_input_reader) :
          d("Inputs not aligned, so consuming and discarding from outer input reader.")
          self.get_and_discard(concrete_input_reader, current_container)
        
        concrete_input_reader = None
      
      # Now, the item could be a container (e.g. a list, dict, or some other
      # AbstractContainer), so to save the _put definition from having to wrap
      # it for stateful consumption of items, let's do it here.
     
      # TODO: We need to check that the container, if from an item, has been fully consumed
      # here and raise an LensException if it has not.
      item_as_container = ContainerFactory.wrap_container(item)
      if has_value(item_as_container) :
        # The item is now represented as a consumable container.
        item = None
        current_container = item_as_container
        # Set the us as the lens of this container, which it will use to determine alignment mode, etc.
        current_container.set_container_lens(self)
      else :
        # When PUTing a non-container item, for consistancy, should cast to string (e.g. if int
        # passed) and discard current container from this branch.
        item = str(item)
        current_container = None


      # Now that arguments are set up, call PUT proper on our lens.
      try :
        output = self._put(item, concrete_input_reader, current_container)

        # Check the container items have been fully consumed by this lens.
        if has_value(item_as_container) and GlobalSettings.check_consumption and not current_container.is_fully_consumed() :
          raise NotFullyConsumedException("The container %s has not been fully consumed." % current_container)
      finally:
        # Now recover the original state of the item, including its meta data,
        # whether put succeeded or not.
        original_item._meta_data = original_meta_data
        if isinstance(original_item, Rollbackable) :
          original_item._set_state(original_state)
        
        
   
    # If instead of an item we have a container, instruct the container to put
    # an item into the lens.  This gives the container much flexibilty about
    # how it chooses an item to PUT, perhaps even doing so tentatively.
    elif has_value(current_container) :
      assert(isinstance(current_container, AbstractContainer))
      output = current_container.consume_and_put_item(self, concrete_input_reader)
    
    # Catch-all case.  We should have been passed an item to PUT.
    else :
      raise LensException("This typed lens expected to PUT an item either directly or from a container.")

    # Report what we PUT.
    if IN_DEBUG_MODE :
      if has_value(output) :
        d("PUT: '%s'" % output)
      else :
        d("PUT: NOTHING")

    # If appropriate, check the input was fully consumed by this lens
    if isinstance(concrete_input, str) and GlobalSettings.check_consumption and not original_concrete_input_reader.is_fully_consumed() :
      raise NotFullyConsumedException("The following input remains to be consumed by this lens: '%s'" % original_concrete_input_reader.get_remaining())

    return output



  def get_and_discard(self, concrete_input, current_container) :
    """
    Sometimes we wish to consume input but discard any items GOTten.
    Note, it is tempting to somehow not use the current_container, though some lenses might
    one day use the current container state, so we must first store items in the
    container before reverting it.  For example, the opening and closing tags in
    XML-like structures.
    """
    # If we have a container, store its start state.
    if has_value(current_container): container_start_state = current_container._get_state()

    # Issue the get.
    self.get(concrete_input, current_container)
    
    # Now revert the state.
    if has_value(current_container): current_container._set_state(container_start_state)

  

  def has_type(self) :
    """Determines if this lens will GET and PUT a variable - a STORE lens."""
    return self.type != None


  def container_get(self, lens, concrete_input_reader, current_container) :
    """
    Convenience function to handle the case where:
      - if there is a container, get and store an item into it
      - if there is no container, call get and check nothing is returned
    
    This simplifies lens such as And and Repeat, whose logic does not have to
    worry about whether or not it is acting as a STORE lens.
    """
    if has_value(current_container) :
      current_container.get_and_store_item(lens, concrete_input_reader)
    else :
      # Call get on lens passing no container, checking it returns no item.
      assert_msg(lens.get(concrete_input_reader, None) == None,
        "The untyped container lens %s did not expect the sub-lens %s to return an item" % (self, lens)
      )

  def container_put(self, lens, concrete_input_reader, current_container) :
    """Reciprocal of container_get."""
    if lens.has_type() :
      assert_msg(has_value(current_container), "Lens %s expected an enclosing container from which to pluck an item." % lens)
      return current_container.consume_and_put_item(lens, concrete_input_reader)
    else :
      # Otherwise, we pass through arguments (e.g. for non-store sublenses or
      # lens that enclose STORE lenses)
      return lens.put(None, concrete_input_reader, current_container)

  def set_sublens(self, sublens) :
    """Used if only a single sublens is required (e.g. the Forward lens)."""
    self.lenses = [self._preprocess_lens(sublens)]

  def extend_sublenses(self, new_sublenses) :
    """
    Adds new sublenses to this lens, being sure to preprocess them (e.g. convert
    strings to Literal lenses, etc.).
    """
    for new_sublens in new_sublenses :
      self.lenses.append(self._preprocess_lens(new_sublens))

  #
  # Helper methods.
  #

  def _normalise_concrete_input(self, concrete_input) :
    """If a string is passed, ensure it is normalised to a ConcreteInputReader."""
    if not has_value(concrete_input) :
      return None
    
    if isinstance(concrete_input, str) :
      concrete_input = ConcreteInputReader(concrete_input)
   
    assert_msg(isinstance(concrete_input, ConcreteInputReader), "Expected to have a ConcreteInputReader not a %s" % type(concrete_input))
    return concrete_input

  def _create_lens_container(self):
    """Creates a container for this lens, if the lens is of a container type."""
    return ContainerFactory.create_container(self.type)


  # XXX: I don't really like these forward declarations, but for now this does
  # the job.  Perhaps lenses can be registered with the framework for more
  # flexible coercion.
  @staticmethod
  def _coerce_to_lens(lens_operand):
    """
    Intelligently converts a type to a lens (e.g. string instance to a Literal
    lens) to ease lens definition; or a class or instance.
    """
    # Coerce string to Literal
    if isinstance(lens_operand, str) :
      lens_operand = Literal(lens_operand)
    
    # Coerce LensObject class to its internally defined lens, such that the lens will GET
    # and PUT instances of that class.
    elif inspect.isclass(lens_operand) and issubclass(lens_operand, LensObject) :
      assert_msg(hasattr(lens_operand, "__lens__"), "LensObject %s defines no __lens__ variable" % lens_operand)
      # Note, we also coerce __lens__ to a lens, just for completeness (e.g. if
      # lens was simply a string, it would be coerced to a Literal lens.
      lens_operand = Group(Lens._coerce_to_lens(lens_operand.__lens__), type=lens_operand)
    
    assert_msg(isinstance(lens_operand, Lens), "Unable to coerce %s to a lens" % lens_operand)
    return lens_operand

  def _preprocess_lens(self, lens) :
    """
    Preprocesses a lens to enable type-to-lens conversion. This will be
    called before processing lens arguments.
    """
    lens = Lens._coerce_to_lens(lens)
    return lens


  #
  # Allow the potential modification of incoming and outgoing items (e.g. to
  # handle auto_list, where a single-item list is converted to and from a
  # single item)
  #

  def _process_outgoing_item(self, item) :
    """
    Allows post-processing of an item in the GET direction.
    
    For example, to handle auto_list, where a single-item list is converted to
    a single item.
    """
    
    # This allows a list singleton to be returned as a single item, for
    # convenience.
    if self.options.auto_list == True and self.has_type() and issubclass(self.type, list) and len(item) == 1:
      # The easy part is extracting a singleton from the list, but we must
      # also preserve the source meta data of the list item by piggybacking it onto
      # the extracted item's meta data

      list_meta_data = item._meta_data
      singleton_meta_data = item[0]._meta_data
      item = item[0]
      item._meta_data = list_meta_data
      item._meta_data.singleton_meta_data = singleton_meta_data
    
    # This allows a list of chars to be combined into a string.
    elif self.options.combine_chars and self.has_type() and issubclass(self.type, list):
      # Note, care should be taken to use this only when a list of single chars is used.
      # XXX: Note, we actually loose each char's meta data here, but this should not be a problem in most cases.
      original_meta = item._meta_data
      item = enable_meta_data("".join(item))
      item._meta_data = original_meta
 
    # Mark if this item is to be used AS a label.
    if self.options.is_label :
      item._meta_data.is_label = True
    # Mark the item to have a static label.
    elif has_value(self.options.label) :
      item._meta_data.label = self.options.label

    return item


  def _process_incoming_item(self, item) :
    """
    Pre-processes an item in the PUT direction.  We can assume we are a lens
    with a type when this is called.
    
    For example, reciprocating the auto_list example in
    _process_outgoing_item.
    """
   
    # TODO: What if item was GOTen from auto list but is not being put directly
    # - it will try to use the wrong source meta.

    # Handle auto_list, expanding an item into a list, being careful to restore any meta data.
    if self.options.auto_list == True and issubclass(self.type, list) and not isinstance(item, list):

      # Create some variables to clarify the process.
      singleton = item
      list_meta_data = item._meta_data
      singleton_meta_data = item._meta_data.singleton_meta_data
      
      # Safeguard to ensure piggybacked data cannot be used more than once.
      list_meta_data.singleton_meta_data = None
      
      # Wrap the singleton in a list, giving it the meta_data of the item.
      item = list_wrapper([singleton])
      item._meta_data = list_meta_data

      # Ensure the singleton has its meta data restored, if it was maintained.
      if singleton_meta_data :
        singleton._meta_data = singleton_meta_data

    # This handles the case where we might have GOTten an item from an auto list
    # but are now putting it directly.  We simply reinstate the singleton meta
    # data on the item.
    elif has_value(item._meta_data.singleton_meta_data) :
      item._meta_data = item._meta_data.singleton_meta_data

    # This allows a list of chars to be combined into a string.
    elif isinstance(item, str) and self.options.combine_chars and issubclass(self.type, list):
      # Note, care should be taken to use this only when a list of single chars is used.
      original_meta = item._meta_data
      item = enable_meta_data(list(item))
      item._meta_data = original_meta

    return item


  #
  # Operator overloads to make for cleaner lens construction.
  #
  
  def __add__(self, other_lens): return And(self, self._preprocess_lens(other_lens))
  def __or__(self, other_lens): return Or(self, self._preprocess_lens(other_lens))

  # Reflected operators, so we can write: lens = "a string" + <some_lens>
  def __radd__(self, other_lens): return And(self._preprocess_lens(other_lens), self)
  def __ror__(self, other_lens): return Or(self._preprocess_lens(other_lens), self)


  #
  # Specialised lenses must override these to implement their GET and PUT
  # proper.
  #

  def _get(self, concrete_input_reader, current_container) :
    """GET proper for a specific lens."""
    raise NotImplementedError("")

  def _put(self, item, concrete_input_reader, current_container) :
    """
    PUT proper for a specific lens.
    
    Note, a low-level lens (i.e. that directly consumes input or tokens)
    should check if it has been mistakenly passed an item to PUT if it is not
    operating as a STORE lens (i.e. if the lens has no type).
    """
    raise NotImplementedError("")


  #
  # For debugging
  #
  
  def _display_id(self) :
    """Useful for identifying specific lenses in debug traces."""
    # If we have a specic name, use it.
    if self.name :
      return self.name

    # If no name, a hash with small range gives us a reasonably easy way to
    # distinguish lenses in debug traces.
    return str(hash(self) % 256)

  # String representation.
  def __str__(self) :
    # Bolt on the class name, to ease debugging.
    return "%s(%s)" % (self.__class__.__name__, self._display_id())
  __repr__ = __str__

  def __instance_name__(self) :
    """Used by my nbdebug module to display a custom debug message context string."""
    return self.name or str(self)



#########################################################
# Core lenses - required fundamental lenses.
#########################################################
    
class And(Lens) :
  """A lens that is formed from the ANDing of two sub-lenses."""

  def __init__(self, *lenses, **options):
   
    # Must always remember to invoke the parent lens, so it can initialise
    # common arguments.
    super(And, self).__init__(**options)

    # Flatten sub-lenses that are also Ands, so we don't have too much nesting,
    # which makes debugging lenses a nightmare.
    for lens in lenses :
      # Note, isinstance would be too vague - Word() was absorbed due to this.
      # Also, self.class == lens.__class__ still too vague -> Collapsed my nested lists!
      if lens.__class__ == And and self.__class__ == And:
        self.extend_sublenses(lens.lenses)
      else :
        self.extend_sublenses([lens])


  def _get(self, concrete_input_reader, current_container) :
    """Sequential GET on each lens."""
    for lens in self.lenses :
      self.container_get(lens, concrete_input_reader, current_container)

    # Important: we should not return anything, since we work on the outer
    # container, that the Lens class sets up for us in Lens.get regardless if our
    # lens created the container or not.


  def _put(self, item, concrete_input_reader, current_container) :
    """Sequential PUT on each lens."""
    # In the same way that we do not return an item in GET, we do not expect
    # to PUT an individual item; again, this is handle in Lens.put
    assert_msg(item == None, "Lens %s did not expect to PUT an individual item %s, since it PUTs from a container" % (self, item))

    # Simply concatenate output from the sub-lenses.
    output = ""
    for lens in self.lenses :
      output += lens.put(None, concrete_input_reader, current_container)

    return output
    

  @staticmethod
  def TESTS() :
    d("GET")
    lens = And(AnyOf(alphas, type=str), AnyOf(nums, type=int), type=list)
    got = lens.get("m0")
    assert(got == ["m", 0])
    
    d("PUT")
    got[0] = "d" # Modify an item, though preserving list source info in meta.
    assert(lens.put(got) == "d0")
 
    d("CREATE") # The new list and items will hold no meta.
    assert(lens.put(["z", 8]) == "z8")

    d("Input alignment test")
    # Now test alignment of input with a more complex lens
    sub_lens = Group(AnyOf(alphas,type=str) + AnyOf("*+", default="*") + AnyOf(nums, type=int), type=list)
    lens = Group(sub_lens + sub_lens, type=list)
    auto_name_lenses(locals()) 
    got = lens.get("a+3x*6")
    assert(got == [["a", 3], ["x", 6]])

    # Now re-order the items
    got.append(got.pop(0))
    output = lens.put(got)
    # And the non-stored input (i.e. '*' and '+') should have been carried with the abstract items.
    assert(output == "x*6a+3")

    # And for CREATE, using default value for non-store lens.
    output = lens.put([["b",9], ["c", 4]])
    assert(output == "b*9c*4")


class Or(Lens) :
  """
  This is the OR of two lenses.
  """
  
  def __init__(self, *lenses, **options):
    super(Or, self).__init__(**options)

    # Flatten sub-lenses that are also Ors, so we don't have too much nesting, which makes debugging lenses a nightmare.
    for lens in lenses :
      # Note, isinstance would be too vague - see my note in And.
      if lens.__class__ == Or and self.__class__ == Or:
        self.extend_sublenses(lens.lenses)
      else :
        self.extend_sublenses([lens])


  def _get(self, concrete_input_reader, current_container) :
    """
    Calls get on each lens until the firstmost succeeds.
    
    Note that the lens should be designed accordingly to break ties over
    multiple valid paths.
    """
    for lens in self.lenses :
      try :
        with automatic_rollback(concrete_input_reader, current_container) :
          return lens.get(concrete_input_reader, current_container)
      except LensException:
        pass
        
    raise LensException("We should have GOT one of the lenses.")

  def _put(self, item, concrete_input_reader, current_container) :
    """
    It is important to realise that here we can either do a:
      - straight PUT, where the lens both consumes input and PUTs an item
      - cross PUT, where one lens consumes input and another PUTS an item

    Also, it is useful to consider a lens l = A | Empty(), since if we first try
    straight PUT with each lens, the Empty lens will always succeed, possibly
    resulting in input not being consumed correctly by lens A.  This case
    influences my redesign of this algorithm.
    """

    # Algorithm
    #
    # For lens_a in lenses
    #   Try straight put with lens_a -> return if succeed
    #   lens_a.get_and_discard(input)
    #
    #   For lens_b in lenses, lens_b != lens_a
    #     lens.put(input=None)

    # Store the initial state.
    initial_state = get_rollbackables_state(concrete_input_reader, current_container)

    for lens_a in self.lenses:
      # Try a straight put on the lens - this will also succeed if there is no
      # input.
      try :
        with automatic_rollback(concrete_input_reader, current_container, initial_state=initial_state) :
          return lens_a.put(item, concrete_input_reader, current_container)
      except LensException:
        pass
     
      # If we have a concrete_input_reader, we will next attempt a cross PUT.
      if not concrete_input_reader :
        continue

        
      # Try to consume input with the lens_a
      try :
        with automatic_rollback(concrete_input_reader, current_container, initial_state=initial_state) :
          lens_a.get_and_discard(concrete_input_reader, current_container)
      except LensException:
        continue

      # If the GET suceeded with lens_a, try to PUT with one of the other
      # lenses.
      for lens_b in self.lenses:
        if lens_a is lens_b:
          continue

        try :
          with automatic_rollback(concrete_input_reader, current_container, initial_state=initial_state) :
            return lens_b.put(item, None, current_container)
        except LensException:
          pass

    raise LensException("We should have PUT one of the lenses.")


  def _display_id(self) :
    """For debugging clarity."""
    return " | ".join([str(lens) for lens in self.lenses])


  @staticmethod
  def TESTS() :
    
    d("GET")
    lens = AnyOf(alphas, type=str) | AnyOf(nums, type=int)
    concrete_input_reader = ConcreteInputReader("abc")
    got = lens.get(concrete_input_reader)
    assert(got == "a" and concrete_input_reader.get_remaining() == "bc")
    concrete_input_reader = ConcreteInputReader("123")
    got = lens.get(concrete_input_reader)
    assert(got == 1 and concrete_input_reader.get_remaining() == "23")

    d("PUT")
    # Test straight put
    concrete_input_reader = ConcreteInputReader("abc")
    assert(lens.put("p", concrete_input_reader) == "p" and concrete_input_reader.get_remaining() == "bc")

    # Test cross put
    concrete_input_reader = ConcreteInputReader("abc")
    assert(lens.put(4, concrete_input_reader) == "4")
    assert(concrete_input_reader.get_remaining() == "bc")
    
    d("Test with default values")
    lens = AnyOf(alphas, type=str) | AnyOf(nums, default=3)
    assert(lens.put() == "3")
    assert(lens.put("r") == "r")
    
    # And whilst we are at it, check the outer lens default overrides the inner lens.
    lens.default = "x"
    assert(lens.put() == "x")

    d("Test sensible handling of Empty lens.")
    # Actually, this issue (if not fixed) will manifest only if we use container.
    lens = AnyOf(alphas, type=str) | Empty()
    concrete_input_reader = ConcreteInputReader("b4")
    assert(lens.put("4", concrete_input_reader) == "" and concrete_input_reader.get_remaining() == "4")

    d("Test sensible handling of Empty lens with container.")
    lens = Group((AnyOf(alphas, type=str) | Empty()) + AnyOf(nums, type=int), type=list)
    concrete_input_reader = ConcreteInputReader("a4")
    got = lens.get(concrete_input_reader)
    assert(got == ["a", 4])
    
    # We expect that: put char that is not in alphas, so Empty used, yet full string consumed.
    # This is an important test that OR works as expected (i.e. according to the
    # ordering of operand lenses) with a lens such as
    # Empty()
    concrete_input_reader.reset()
    del got[0] # Important that we use 'got' so input is aligned.
    assert(lens.put(got, concrete_input_reader) == "4")
    assert(concrete_input_reader.is_fully_consumed())
    

    


class AnyOf(Lens) :
  """
  The first useful low-level lens. Matches a single char within a specified
  set, and can also be negated.
  """
  
  def __init__(self, valid_chars, negate=False, **options):
    super(AnyOf, self).__init__(**options)
    self.valid_chars, self.negate = valid_chars, negate
 
  def _get(self, concrete_input_reader, current_container) :
    """
    Consumes a valid char from the input, returning it if we are a STORE
    lens.
    """
    char = None
    try:
      char = concrete_input_reader.consume_char()
      if not self._is_valid_char(char) :
        raise LensException("Expected char %s but got '%s'" % (self._display_id(), truncate(char)))
    except EndOfStringException:
      raise LensException("Expected char %s but at end of string" % (self._display_id()))
   
    if self.has_type() :
      return char
    else :
      return None


  def _put(self, item, concrete_input_reader, current_container) :
    """
    If a store lens, tries to output the given char; otherwise outputs
    original char from concrete input.
    """
    # If we are not a store lens, simply return what we would consume from the input.
    if not self.has_type() :
      # We should not have been passed an item.
      assert(not has_value(item))
      if has_value(concrete_input_reader) :
        concrete_start_position = concrete_input_reader.get_pos()
        self._get(concrete_input_reader, current_container)
        return concrete_input_reader.get_consumed_string(concrete_start_position)
        
      else :
        raise NoDefaultException("Cannot CREATE: a default should have been set on lens %s, or a higher lens." % self)
    
    # If this is PUT (vs CREATE) then first consume input.
    if concrete_input_reader :
      self.get(concrete_input_reader)
    
    
    if not (isinstance(item, str) and len(item) == 1 and self._is_valid_char(item)) :
      raise LensException("Invalid item '%s', expected %s." % (item, self._display_id()))
    return item


  def _is_valid_char(self, char) :
    """Tests if that passed is a valid character for this lens."""
    if self.negate :
      return char not in self.valid_chars
    else :
      return char in self.valid_chars

  def _display_id(self) :
    """To aid debugging."""
    if self.name :
      return self.name
    if self.negate :
      return "not in [%s]" % range_truncate(self.valid_chars)
    else :
      return "in [%s]" % range_truncate(self.valid_chars)
  

  @staticmethod
  def TESTS() :
   
    lens = AnyOf(alphas, type=str, some_property="some_val")
    assert(lens.options.some_property == "some_val") # Might as well test lens options working.
    
    d("GET")
    got = lens.get("m")
    assert(got == "m")

    # Test putting back what we got (i.e. with meta)
    d("PUT")
    assert(lens.put(got) == "m")

    d("CREATE")
    output = lens.put("d")
    assert(output == "d")
   
    d("Test default of non-typed lens is created")
    lens = AnyOf(alphas, default="x")
    output = lens.put()
    assert(output == "x")

    d("Test failure when default value required.")
    lens = AnyOf(alphas)
    with assert_raises(NoDefaultException) :
      lens.put()
    
    d("TEST type coercion")
    lens = AnyOf(nums, type=int)
    assert(lens.get("3") == 3)
    assert(lens.put(8) == "8")
    
    # Check default converted to string.
    lens = AnyOf(nums, default=5)
    assert(lens.put() == "5")
    

class Repeat(Lens) :
  """
  Applies a repetition of the givien lens (i.e. kleene-star).
  """

  def __init__(self, lens, min_count=1, max_count=None, **options):
    """
    Arguments:
      lens - the lens to repeat
      min_count - the min repetitions
      max_count - maximum repetitions (must be > 0 if set)
    """
    super(Repeat, self).__init__(**options)
    assert(min_count >= 0)
    if has_value(max_count) :
      assert(max_count > min_count)
    
    self.min_count, self.max_count = min_count, max_count
    self.extend_sublenses([lens])


  def _get(self, concrete_input_reader, current_container) :
    """Calls a sequence of GETs on the sub-lens."""

    # Algorithm
    #
    # Loop until max count reached or lens fails to alter input state (which
    # could happen indefinitely.

    # For brevity.
    lens = self.lenses[0]
   
    # For tracking how many successful GETs
    no_got = 0

    while(True) :
      # Instantiate the rollback context, so we can later check if any state was changed.
      rollback_context = automatic_rollback(concrete_input_reader, current_container, check_for_state_change=True)
      try :
        with rollback_context :
          self.container_get(lens, concrete_input_reader, current_container)
        
        # If the lens changed no state, then we must break, otherwise continue
        # for ever.
        if not rollback_context.some_state_changed :
          d("Lens %s changed no state during this iteration, so we must break out - or spin for ever" % lens)
          break
        
        no_got += 1

        # Don't get more than maximim
        if has_value(self.max_count) and no_got == self.max_count :
          break
      except LensException :
        break

    if no_got < self.min_count :
      raise TooFewIterationsException("Expected at least %s successful GETs but got only %s" % (self.min_count, no_got))
    

  def _put(self, item, concrete_input_reader, current_container) :
    """Calls a sequence of PUTs on the sub-lens."""

    # Algorithm
    #
    # - First we try to PUT with the lenses (first using input if it is available)
    # - Then, we must check that the minumum items have been consumed from input:
    # we may have PUT all new items (i.e. without input consumption) so still
    # need to call GET for a minimum number of times.
    # - In both cases we must break out if a lens changes no state.

    # For brevity.
    lens = self.lenses[0]

    no_got = 0    # For checking how many items of input were consumed
    no_put = 0    # For checking how many items were PUT.
    output = ""

    # This simplifies our algorithm.
    if concrete_input_reader :
      input_readers = [concrete_input_reader, None]
    else :
      input_readers = [None]

    #
    # Handle the PUT/CREATEs
    #

    for input_reader in input_readers :
      
      # Allows the while loop to request breakout from outer for loop.
      break_for_loop = False

      while True :
        # Call PUT on the lens and break this while loop if no state changed or we
        # get a LensException.  Also, break the for loop if we PUT max count.
        rollback_context = automatic_rollback(input_reader, current_container, check_for_state_change=True)
        try :
          with rollback_context:
            put = self.container_put(lens, input_reader, current_container)
        except LensException:
          # TODO: To support deletion (i.e. when no item matches this input, wrap lens as: lens | Empty()
          # Infact we should not expect a LensException - only break out when no state changes.
          break

        if not rollback_context.some_state_changed :
          d("Lens %s changed no state during this iteration, so we must break out - or spin for ever" % lens)
          break

        output += put
        no_put += 1
        
        # If the lens succeeded when we used an input reader, we assume we consumed
        # input with the lens.
        if has_value(input_reader) :
          no_got += 1

        # We have PUT enough items now.
        if no_put == self.max_count :
          d("We have put a maximum number of items now, so breaking out.")
          break_for_loop = True
          break
      
      if break_for_loop :
        break

    
    #
    # Now consume and discard any remaining input if necessary (i.e. if we put
    # fewer than we had originally).  This is basically the same implentation of
    # get, though we discard any items and continue to track no_got for the puts above.
    #

    if concrete_input_reader and no_got < self.max_count:
      
      d("Now consuming and discarding excess input.")
      
      # Iterate over the input with our lens, consuming as much of it as
      # possible.
      while(True) :
        # Instantiate the rollback context, so we can later check if any state was changed.
        rollback_context = automatic_rollback(concrete_input_reader, current_container, check_for_state_change=True)
        try :
          with rollback_context :
            # XXX: Inefficient to discard container items each time.
            lens.get_and_discard(concrete_input_reader, current_container)
          
          # If the lens changed no state, then we must break, otherwise continue
          # for ever.
          if not rollback_context.some_state_changed :
            d("Lens %s changed no state during this iteration, so we must break out - or spin for ever" % lens)
            break
          
          no_got += 1

          # Don't get more than maximim
          if has_value(self.max_count) and no_got == self.max_count :
            break
        except LensException :
          break
      


    if no_put < self.min_count :
      raise TooFewIterationsException("Expected at least %s successful PUTs but put only %s" % (self.min_count, no_put))
  
    # Sanity check.
    if concrete_input_reader:
      # This should not happen... I think.
      assert(no_got >= self.min_count)

    return output


  @staticmethod
  def TESTS() :
   
    # Note, for some of these tests we need to ensure we PUT rather than CREATE
    # the lists so we can flex the code when input is aligned.

    lens = Repeat(AnyOf(nums, type=int), min_count=3, max_count=5, type=list)
    d("GET")
    assert(lens.get("1234") == [1,2,3,4])
    with assert_raises(TooFewIterationsException) :
      lens.get("12")
    # Test max_count
    assert(lens.get(ConcreteInputReader("12345678")) == [1,2,3,4,5])

    test_description("Put as many as were there originally.")
    input_reader = ConcreteInputReader("98765")
    assert(lens.put([1,2,3,4,5], input_reader) == "12345" and input_reader.get_remaining() == "")
    
    test_description("Put a maximum (5) of the items (6)")
    input_reader = ConcreteInputReader("987654321")
    GlobalSettings.check_consumption = False
    assert(lens.put([1,2,3,4,5,6], input_reader) == "12345" and input_reader.get_remaining() == "4321")
    GlobalSettings.check_consumption = True
    
    test_description("Put more than there were originally.")
    input_reader = ConcreteInputReader("981abc")
    got = lens.get(input_reader)
    got.insert(2, 3)
    input_reader.reset()
    assert_equal(lens.put(got, input_reader), "9831")
    assert_equal(input_reader.get_remaining(), "abc")

    test_description("PUT fewer than got originally, but consume up to max from the input.")
    input_reader = ConcreteInputReader("87654321")
    got = lens.get(input_reader)
    del got[2]   # Remove the 6
    input_reader.reset()
    assert(lens.put(got, input_reader) == "8754")
    assert(input_reader.get_remaining() == "321")

    test_description("Test non-typed lenses.")
    lens = Repeat(AnyOf(nums, default=8))
    input_reader = ConcreteInputReader("12345abc")
    d(lens.get(input_reader) == None and input_reader.get_remaining() == "abc")
    input_reader.reset()
    # Lens should make use of outer input, since not supplied by an item.
    assert(lens.put(None, input_reader, None) == "12345" and input_reader.get_remaining() == "abc")

    test_description("Test the functionality without default values.")
    # Should fail, since lens has no default, so could put infinite items.
    with assert_raises(LensException) :
      lens.put()

    test_description("Test the functionality with default value on Repeat.")
    lens = Repeat(AnyOf(nums), default="54321")
    assert(lens.put() == "54321")
    

    d("Test putting back what we got (i.e. with source meta)")
    lens = Repeat(AnyOf(nums, type=int), type=list)
    assert(lens.put(lens.get("1234")) == "1234")

    d("Test for completeness")
    GlobalSettings.check_consumption = False
    lens = Repeat(AnyOf(nums, type=int), type=list, min_count=0, max_count=1)
    assert(lens.get("abc") == []) # No exception thrown since min_count == 0
    assert(lens.get("123abc") == [1])  # Since max_count == 1
    assert(lens.put([1,2,3]) == "1") # Since max_count == 1

    d("Test combine_chars")
    lens = Repeat(AnyOf(alphas, type=str), type=list, combine_chars=True)
    assert(lens.get("abc123") == "abc")
    assert(lens.put("xyz") == "xyz")

    GlobalSettings.check_consumption = True

    d("Test infinity problem")
    lens = Repeat(Empty(), min_count=3, max_count=None)
    # Will fail to get anything since Empty lens changes no state.
    with assert_raises(LensException) :
      lens.get("anything")
    # Likewise.
    with assert_raises(LensException) :
      lens.put(None)
    
    d("Test the functionality with default value on sub-lens.")
    lens = Repeat(AnyOf(nums, default=4))
    # Should faile since no input or items are consumed by the lens.
    with assert_raises(LensException) :
      lens.put()


class Empty(Lens) :
  """
  Matches the empty string, used by Optional().  Can also set modes for special
  empty matches (e.g. at the start or end of a string).
  """
  
  # Useful modifiers for empty matches.
  START_OF_TEXT = "START_OF_TEXT"
  END_OF_TEXT   = "END_OF_TEXT"

  def __init__(self, mode=None, **options):
    super(Empty, self).__init__(**options)
    self.default = ""
    self.mode = mode


  def _get(self, concrete_input_reader, current_container) :
    
    # Check for special modes.
    if self.mode == self.START_OF_TEXT :
      if concrete_input_reader.get_pos() != 0 :
        raise LensException("Will match only at start of text.")
    elif self.mode == self.END_OF_TEXT :
      if not concrete_input_reader.is_fully_consumed() :
        raise LensException("Will match only at end of text.")

    # Note that, useless as it is, this is actually an item that could potentially be stored that we
    # return, which is why we must explicitly check for None elsewhere in the
    # framework (e.g. use has_value(...)), since "" == False but "" != None.
    if self.has_type() :
      return ""
    return None


  def _put(self, item, concrete_input_reader, current_container) :
    
    if self.has_type() :
      if not (has_value(item) and isinstance(item, str) and item == "") :
        raise LensException("Expected to PUT an empty string")
    else :  
      # Should not have been passed an item.
      if has_value(item) :
        raise LensException("Did not expect a non-store lens to be passed an item")
        
      # We could be called if there is concrete input, such that our default did not intercept.
      assert_msg(has_value(concrete_input_reader), "_put() should never be called on a non-store Empty lens without concrete input.")
      
    # Here goes nothing!
    return ""


  @staticmethod
  def TESTS() :
    
    GlobalSettings.check_consumption = False
    
    d("Test with type")
    lens = Empty(type=str)
    assert(lens.get("anything") == "")
    assert(lens.put("", "anything") == "")
    with assert_raises(LensException) :
      lens.put(" ", "anything")
    
    assert(lens.put("") == "")


    d("Test without type")
    lens = Empty()
    assert lens.get("anything") == None 
    assert lens.put() == ""
    # Lens does not expect to put an item, valid or otherwise.
    with assert_raises(LensException) :
      lens.put("", "anything")
 
    d("Test special modes.")
    lens = Empty(mode=Empty.START_OF_TEXT)
    concrete_reader = ConcreteInputReader("hello")
    # Progress the input reader so lens does not match.
    concrete_reader.consume_char()
    with assert_raises(LensException) : 
      lens.get(concrete_reader)
    
    lens = Empty(mode=Empty.END_OF_TEXT)
    concrete_reader = ConcreteInputReader("h")
    # This should throw an Exception
    with assert_raises(LensException) : 
      lens.get(concrete_reader)
    concrete_reader.consume_char()
    # This should succeed quietly.
    lens.get(concrete_reader)



class Group(Lens) :
  """
  A convenience lens that thinly wraps any lens, basically to set a type.
  Usually this is used to close off a lenses container.
  """

  def __init__(self, lens, **options):
    super(Group, self).__init__(**options)
    assert_msg(self.has_type(), "To be meaningful, you must set a type on %s" % self)
    self.extend_sublenses([lens])

  def _get(self, concrete_input_reader, current_container) :
    return self.lenses[0].get(concrete_input_reader, current_container)

  def _put(self, item, concrete_input_reader, current_container) :
    return self.lenses[0].put(item, concrete_input_reader, current_container)

  @staticmethod
  def TESTS() :
    GlobalSettings.check_consumption = False

    d("GET")
    lens = Group(AnyOf(alphas,type=str) + AnyOf(nums, type=int), type=list)
    got = lens.get("a2b3")
    d(got)
    assert(got == ["a", 2])

    d("PUT")
    assert(lens.put(got, "n6b3") == "a2")

    d("CREATE")
    assert(lens.put(["x", 4]) == "x4")

    d("TEST erroneous Group with no type")
    with assert_raises(AssertionError) :
      lens = Group(AnyOf(nums))
G = Group

class Literal(Lens) :
  """
  A lens that deals with a constant string, usually that will not be stored.
  """
  
  # TODO: Add case insensitivity.

  def __init__(self, literal_string, **options):
    assert(isinstance(literal_string, str) and len(literal_string) > 0)
    super(Literal, self).__init__(**options)
    self.literal_string = literal_string
    if not self.has_type() :
      self.default = self.literal_string
  
  def _get(self, concrete_input_reader, current_container) :
    """
    Consumes a valid char form the input, returning it if we are a STORE
    lens.
    """
    input_string = None
    try:
      input_string = concrete_input_reader.consume_string(len(self.literal_string))
      if input_string != self.literal_string :
        raise LensException("Expected the literal '%s' but got '%s'." % (escape_for_display(self.literal_string), escape_for_display(input_string)))
    except EndOfStringException:
      raise LensException("Expected literal '%s' but at end of string." % (escape_for_display(self.literal_string)))
   
    if self.has_type() :
      return input_string
    else :
      return None


  def _put(self, item, concrete_input_reader, current_container) :
    """
    If a store lens, tries to output the given char; otherwise outputs
    original char from concrete input.
    """
    # If we are not a store lens, simply return what we would consume from the input.
    if not self.has_type() :
      # We should not have been passed an item.
      assert_msg(not has_value(item), "%s did not expected to be passed an item - is a non-store lens" % self)
      if has_value(concrete_input_reader) :
        concrete_start_position = concrete_input_reader.get_pos()
        self._get(concrete_input_reader, current_container)
        return concrete_input_reader.get_consumed_string(concrete_start_position)
        
      else :
        raise NoDefaultException("Cannot CREATE: a default should have been set on lens %s, or a higher lens." % self)
    
    # If this is PUT (vs CREATE) then first consume input.
    if concrete_input_reader :
      self.get(concrete_input_reader)
    
    if item != self.literal_string :
      raise LensException("%s can not PUT %s." % (self, item))
    
    return item


  def _display_id(self) :
    """To aid debugging."""
    # Name is only set after Lens constructor called.
    if hasattr(self, "name") and has_value(self.name) :
      return self.name
    return "'%s'" % escape_for_display(self.literal_string)
  


  @staticmethod
  def TESTS() :
    d("GET")
    lens = Literal("xyz")
    concrete_reader = ConcreteInputReader("xyzabc")
    assert(lens.get(concrete_reader) == None and concrete_reader.get_remaining() == "abc")
    d("PUT")
    assert(lens.put(None) == "xyz")
  
    # XXX: Need to think more about this, and what it impacts.
    # Should flag that we mistakenly passed an item to a non-store low-level
    # lens that could not possibly us it.
    #with assert_raises(LensException) :
    #  lens.put("xyz")

    d("Test as STORE lens, pointless as it is with this lens.")
    lens = Literal("xyz", type=str)
    concrete_reader = ConcreteInputReader("xyzabc")
    assert(lens.get(concrete_reader) == "xyz" and concrete_reader.get_remaining() == "abc")
    concrete_reader = ConcreteInputReader("xyzabc")
    assert(lens.put("xyz", concrete_reader) == "xyz" and concrete_reader.get_remaining() == "abc")





#
# Copyright (c) 2010-2011, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Author: Nick Blundell <blundeln@gmail.com>
# Organisation: www.nickblundell.org.uk
#
"""
Main API functions for using pylens.
"""

# Imports all lenses, some indirectly.
from util_lenses import *


##################################
# High-level API functions
##################################

# Hmmm, do we really need these.

# COuld write Person.get() -> person and person.put()

def get(lens, *args, **kargs) :
  """
  Extracts a python structure from some string structure using the given
  lens, ensuring that the lens parameter is conveniently coerced to an
  appropriate Lens class.

  Example::
    
    get(Person, "Person::name=nick,surname=blundell") -> instance of Person class.
  """
  lens = Lens._coerce_to_lens(lens)
  # Wrap in AutoGroup, so outer container may be ommitted for convenience
  # (usually of testing lens fragments)
  if not lens.has_type() :
    lens = AutoGroup(lens)
  return lens.get(*args, **kargs)

def put(lens_or_instance, *args, **kargs) :
  """
  Puts some python structure back into some string structure.

  Example: put(some_lens, {"a":1, "c":4}) -> "a=1,c=4"
  """
  # If we have an instance of a class which defines its own lens...
  if isinstance(lens_or_instance, LensObject) : #and hasattr(lens_or_instance, "__lens__") :
    assert_msg(hasattr(lens_or_instance, "__lens__"), "LensObject %s defines no __lens__" % lens_or_instance)
    lens = Lens._coerce_to_lens(lens_or_instance.__class__)
    instance = lens_or_instance # For clarity.
    return lens.put(instance, *args, **kargs)
  
  # Otherwise...
  lens = Lens._coerce_to_lens(lens_or_instance)
  # Wrap in AutoGroup, so outer container may be ommitted for convenience
  # (usually of testing lens fragments). We assume above that instance will be
  # wrapped in am appropriately typed group, so only do this here.
  if not lens.has_type() :
    lens = AutoGroup(lens)
  return lens.put(*args, **kargs)


###########################
# Main.
#

def main() :
  # This can be useful for quick testing.
  d("Testing")

if __name__ == "__main__":
  main()



#
# Copyright (c) 2010-2011, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
#
#
# Author: Nick Blundell <blundeln [AT] gmail [DOT] com>
# Organisation: www.nickblundell.org.uk
# 
# Description:
#  Classes and functions relating to abstract items, specically allowing meta
#  data to be added to native python types for tracking an item's concrete
#  origin. 
# 
#

from debug import *
from util import *

META_ATTRIBUTE = "_meta_data"

#
# Wrappers for simple types, so we can transparently add arbitrary properies.
#
class str_wrapper(str) :pass
class int_wrapper(int) :pass
class float_wrapper(float) :pass
class list_wrapper(list) :pass
class dict_wrapper(dict) :pass

def item_has_meta(item) :
  return hasattr(item, META_ATTRIBUTE) 

def enable_meta_data(item) :
  """
  If not already present, this adds a flexible Properties attribute to any
  object for storing meta data (e.g. information about the concrete origin of
  an extracted item).
  
  Note that, since some builtin python types cannot hold
  arbitrary attributes, we wrap them thinly in appropriate classes.
  """  
  assert(has_value(item))

  if not item_has_meta(item) : 
    
    # Wrap simple types to allow attributes to be added to them.
    if isinstance(item, str) : item = str_wrapper(item)
    elif isinstance(item, float) : item = float_wrapper(item)
    elif isinstance(item, int) : item = int_wrapper(item)
    elif isinstance(item, list) : item = list_wrapper(item)
    elif isinstance(item, dict) : item = dict_wrapper(item)
   
    setattr(item, META_ATTRIBUTE, Properties())
  
  return item


#
# TESTS
#

def item_meta_test() :
  d("Started")
  item = "hello"
  item = enable_meta_data(item)

  # Should be able to add any attribute.
  item._meta_data.monkeys = True
  assert(item._meta_data.monkeys == True)
  assert(item._meta_data.bananas == None)



#
# Copyright (c) 2010, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# Author: Nick Blundell <blundeln [AT] gmail [DOT] com>
# Organisation: www.nickblundell.org.uk
# 
# Description:
#   Utilities of global use.
#
import copy
from debug import *


class Properties(object):
  """
  A useful class for holding properties (e.g. meta data or lens options), which
  uses __getattr__ to save us from always asking if it has a particular attribute.
  """
  def __init__(self, **kargs) :
    self.__dict__.update(kargs)

  def __getattr__(self, name) :
    # This is important, since obj.__dict__ would equal None!
    if name.startswith("__") :
      return super(object, self).__getattr__(name)

    if name in self.__dict__ :
      return self.__dict__[name]
    else :
      return None

  def copy(self) :
    return copy.copy(self)

  def unwrap(self) :
    return self.__dict__

  def clear(self) :
    """Useful in test cases."""
    self.__dict__ = {}

  def __str__(self) :
    return str(self.__dict__)
  __repr__ = __str__

  @staticmethod
  def TESTS() :
    d("Testing")
    properties = Properties(food="cheese")
    assert(properties.food == "cheese")
    properties.something = [1,2,3]
    assert(properties.something == [1,2,3])
    assert(properties.nothing == None)


def get_class_attr(obj, name, default=None):
  """Specifically get an attribute of an object's class."""
  return getattr(obj.__class__, name, default)

def get_instance_attr(obj, name, default=None) :
  """
  Specifically get an attribute of an instance (i.e. do not fall back on a
  class attribute with the same name, as does getattr).
  """
  class_value = get_class_attr(obj, name)
  obj_value = getattr(obj, name, default)
  
  # If there was no instance attribute by that name, getattr will return the
  # class_value if it exists, which we do not want.
  if has_value(class_value) and obj_value is class_value :
    return default
  
  return obj_value


def escape_for_display(s) :
  """Substitute certain chars to assist debug traces."""
  if len(s) == 0 :
    return "[EMPTY]"
  return s.replace("\n","[NL]").replace("\t","[TAB]") #.replace(" ","[SP]") # Escape newlines so not to confuse debug output.

def truncate(s, max_len=10) :
  """Truncates a long string so is suitable for display."""
  MAX_LEN = max_len
  display_string = escape_for_display(s)
  if len(s) == 0 :
    return display_string # Display empty string token.
  if len(s) > MAX_LEN :
    display_string = display_string[0:MAX_LEN] + "..."
  return display_string

def range_truncate(s, max_len=8) :
  """Truncate a string to assist debug traces."""
  if len(s) > max_len :
    return s[0:2] + "..." + s[-2:]
  return s


def has_value(var) :
  """To avoid possible comparison bugs with empty values vs None."""
  return var != None


#
# TESTS
#

def attr_test():
  test_description("Test behaviour of getattr when there is ambiguity over instance and class attributes")
  class A:
    a1 = 2
  a = A()

  assert_equal(getattr(a,"a1",None), 2)
  assert_equal(get_instance_attr(a,"a1",None), None)
  a.a1 = 5
  assert_equal(getattr(a,"a1",None), 5)
  assert_equal(get_instance_attr(a,"a1",None), 5)
  assert_equal(get_class_attr(a,"a1",None), 2)






#
# Copyright (c) 2010-2011, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
#
#
# Author: Nick Blundell <blundeln [AT] gmail [DOT] com>
# Organisation: www.nickblundell.org.uk
# 
# Description:
# 
#

class GlobalSettings:
  """
  These are some global settings that affect the functionality of the
  framework.
  """
  
  """
  Check that the (outer most) lens fully consumes the
  input string and that containers are fully consumed in the PUT direction,
  where is possible to know such a thing.
  You might wish to set this to False when developing or debugging your own lenses.
  """
  check_consumption = True



#
# Copyright (c) 2010-2011, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# Author: Nick Blundell <blundeln [AT] gmail [DOT] com>
# Organisation: www.nickblundell.org.uk
# 
#
import inspect
from nbdebug import d, breakpoint, set_indent_function, IN_DEBUG_MODE
from exceptions import *
from containers import *
from readers import *
from util import *
from debug import *
from base_lenses import *
from core_lenses import *


class OneOrMore(Repeat) :
  def __init__(self, *args, **options):
    if "min_count" not in options :
      options["min_count"] = 1
    # Mental note: Don't accidentally write something like super(Repeat...
    super(OneOrMore, self).__init__(*args, **options)

  @staticmethod
  def TESTS() :
    # This is really just to check the lens construction.
    lens = OneOrMore(AnyOf(nums, type=int), type=list)
    assert(lens.get("123") == [1,2,3])
OM = OneOrMore


class ZeroOrMore(Repeat) :
  def __init__(self, *args, **options):
    if "min_count" not in options :
      options["min_count"] = 0
    super(ZeroOrMore, self).__init__(*args, **options)
ZM = ZeroOrMore

class Optional(Or) :
  def __init__(self, lens, **options):
    super(Optional, self).__init__(lens, Empty(), **options)

  @staticmethod
  def TESTS():
    GlobalSettings.check_consumption = False
    lens = Optional(AnyOf(alphas, type=str))
    assert(lens.get("abc") == "a")
    assert(lens.get("123") == None)
    assert(lens.put("a") == "a")
    assert(lens.put(1) == "")
O = Optional

class List(And) :
  """Shortcut lens for delimited lists."""
  def __init__(self, lens, delimiter_lens, **options):
    super(List, self).__init__(lens, ZeroOrMore(And(delimiter_lens, lens)), **options)

  @staticmethod
  def TESTS() :
    
    lens = List(AnyOf(nums, type=int), ",", type=list)
    d("GET")
    assert(lens.get("1,2,3") == [1,2,3])
    d("PUT")
    assert(lens.put([6,2,6,7,4,8]) == "6,2,6,7,4,8")

    # It was getting flattened due to And within And!
    test_description("Test a bug I found with nested lists.")
    INPUT = "1|2,3|4,5|6"
    lens = List(
             List(AnyOf(nums, type=int), "|", name="inner_list", type=list),
             ",", name="outer_list",
             type=list,
           )
    got = lens.get(INPUT)
    assert_equal(got, [[1,2],[3,4],[5,6]])
    got.insert(2, [6,7])
    assert_equal(lens.put(got), "1|2,3|4,6|7,5|6")


class NewLine(Or) :
  """Matches a newline char or the end of text, so extends the Or lens."""
  def __init__(self, **options) :
    super(NewLine, self).__init__("\n", Empty(mode=Empty.END_OF_TEXT), **options)

  # TODO: Ensure it puts a \n regardless of being at end of file, to allow
  # appending. Could hook put

  @staticmethod
  def TESTS() :
    lens = NewLine()
    assert(lens.get("\n") == None)
    assert(lens.get("") == None)
    with assert_raises(LensException) :
      lens.get("abc")
    assert(lens.put("\n") == "\n")

NL = NewLine # Abbreviation

class Word(And) :
  """
  Useful for handling keywords of a specific char range.
  """
  def __init__(self, body_chars, init_chars=None, min_count=1, max_count=None, negate=False, **options):

    assert_msg(min_count > 0, "min_count should be more than zero.")

    # For convenience, enable type if label or is_label is set on this lens.
    if "is_label" in options or "label" in options :
      options["type"] = str

    if "type" in options and has_value(options["type"]):
      assert_msg(options["type"] == str, "If set the type of Word should be str.")
      any_of_type = str
      # Ensure the And type is list
      options["type"] = list
    else :
      any_of_type = None
      and_type = None

    # Ensure chars are combined if this is a STORE lens.
    options["combine_chars"] = True

    left_lens = AnyOf(init_chars or body_chars, type=any_of_type)
    right_lens = Repeat(AnyOf(body_chars, type=any_of_type), min_count=min_count-1, max_count=max_count and max_count-1 or None)
    
    super(Word, self).__init__(left_lens, right_lens, **options)

  @staticmethod
  def TESTS() :

    GlobalSettings.check_consumption = False
    
    lens = Word(alphanums, init_chars=alphas, type=str, max_count=5)
    d("GET")
    assert(lens.get("w23dffdf3") == "w23df")
    with assert_raises(LensException) :
      assert(lens.get("1w23dffdf3") == "w23df")

    d("PUT")
    assert(lens.put("R2D2") == "R2D2")
    
    with assert_raises(LensException) :
      lens.put("2234") == "R2D2"
    
    # XXX: Should fail if length checking working correctly.
    #with assert_raises(LensException) :
    #  lens.put("TooL0ng")
 
    
    d("Test with no type")
    lens = Word(alphanums, init_chars=alphas, max_count=5, default="a123d")
    assert(lens.get("w23dffdf3") == None)
    concrete_input_reader = ConcreteInputReader("ab12_3456")
    assert(lens.put(None, concrete_input_reader) == "ab12")
    assert(concrete_input_reader.get_remaining() == "_3456")
    assert(lens.put() == "a123d")



class Whitespace(Or) :
  """
  Whitespace helper lens, that knows how to handle (logically) continued lines with '\\n'
  or that preclude an indent which are useful for certain config files.
  """
  
  def __init__(self, default=" ", optional=False, space_chars=" \t", slash_continuation=False, indent_continuation=False, **options):
    # Ensure default gets passed up to parent class - we use default to
    # determine if this lens is optional

    if "type" in options and has_value(options["type"]):
      # XXX: Could adapt this for storing spaces, though to be useful would need
      # to construct in such a way as to combine chars.
      assert_msg(False, "This lens cannot be used as a STORE lens")
      
    # XXX: This could be used later when we wish to make this a STORE lens.
    word_type = None

    # TODO: Could also use default to switch on, say, indent_continuation.

    # Set-up a lens the literally matches space.
    spaces = Word(space_chars, type=word_type, name="spaces")
    
    or_lenses = []
    
    # Optionally, augment with a slash continuation lens.
    if slash_continuation :
      or_lenses.append(Optional(spaces) + "\\\n" + Optional(spaces))


    # Optionally, augment with a indent continuation lens.
    if indent_continuation :
      or_lenses.append(Optional(spaces) + "\n" + spaces)

    # Lastly, add the straighforward spaces lens - since otherwise this would match before the others.
    or_lenses.append(spaces)

    # If the user set the default as the empty space, the Empty must also be a valid lens.
    if default == "" or optional:
      or_lenses.append(Empty())

    # Set up options for Or.
    options["default"] = default
    super(Whitespace, self).__init__(*or_lenses, **options)

  @staticmethod
  def TESTS() :
   
    GlobalSettings.check_consumption = False

    # Simple whitespace.
    lens = Whitespace(" ")
    concrete_input_reader = ConcreteInputReader("  \t  xyz")
    assert(lens.get(concrete_input_reader) == None and concrete_input_reader.get_remaining() == "xyz")
    assert(lens.put() == " ")
    
    # Test that the Empty lens is valid when the default space is set to empty string (i.e. not space).
    lens = Whitespace("")
    assert(lens.get("xyz") == None)
    assert(lens.put() == "")

    # With slash continuation.
    lens = Whitespace(" ", slash_continuation=True)
    concrete_input_reader = ConcreteInputReader("  \t\\\n  xyz")
    assert(lens.get(concrete_input_reader) == None and concrete_input_reader.get_remaining() == "xyz")

    # With indent continuation.
    lens = Whitespace(" ", indent_continuation=True)
    concrete_input_reader = ConcreteInputReader("  \n xyz")
    assert(lens.get(concrete_input_reader) == None and concrete_input_reader.get_remaining() == "xyz")

WS = Whitespace # Abreviation.


class NullLens(Lens) :
  """
  When writing new lenses, particularly in a top-down fashion, this lens is
  useful for filling in lens branches that are yet to be completed.
  """
  def _get(self, concrete_input_reader) :
    raise LensException("NullLens always fails, and is useful as a filler for the incremental writing of lenses.")
  def _put(self, abstract_token, concrete_input_reader) :
    raise LensException("NullLens always fails, and is useful as a filler for the incremental writing of lenses.")

  # He, he. I won't test this one.

class KeyValue(Group) :
  """
  Simply sets up the Group as an auto_list, which is useful when we just wish
  to store a value by a key.
  """
  def __init__(self, *args, **options):
    if "type" not in options:
      options["type"] = list
    if "auto_list" not in options:
      options["auto_list"] = True
    super(KeyValue, self).__init__(*args, **options)

class BlankLine(And) :
  """
  Matches a blank line (i.e. optional whitespace followed by NewLine().
  """
  def __init__(self, **options):
    super(BlankLine, self).__init__(WS(""), NewLine(), **options)


class Keyword(Word) :
  """
  A lens for matching a typical keyword.
  """
  def __init__(self, additional_chars="_", **options):
    super(Keyword, self).__init__(alphanums+additional_chars, init_chars = alphas+additional_chars, **options)


class AutoGroup(Group):
  """
  Sometimes it may be convenient to not explicitly set a type on an outer lens
  in order to extract one or more items from sub-lenses, so this lens allows an
  outer container to be set automatically, using auto_list such that a single
  item may be passed through the lens.  If the enclosed lens has a type, then
  this lens simply becomes a transparent wrapper.
  """
 
  def __init__(self, lens, **options):
    """Note, this replaces __init__ of Group, which checks for a type."""
    if not lens.has_type() :
      options["type"] = list
      options["auto_list"] = True
    super(Group, self).__init__(**options)
    self.extend_sublenses([lens])


class HashComment(And) :
  """A common hash comment."""
  def __init__(self, **options):
    super(HashComment, self).__init__("#", Until(NewLine()), NewLine(), **options)



#
# Copyright (c) 2010, Nick Blundell
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Nick Blundell nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
#
#
# Author: Nick Blundell <blundeln [AT] gmail [DOT] com>
# Organisation: www.nickblundell.org.uk
#
# Description:
#  Consumable containers for abstract storage of model items.
#
import inspect
import copy
import re

from debug import *
from exceptions import *
from util import *
from item import *
from rollback import *

# Item alignment modes for containers.
SOURCE = "SOURCE"
MODEL = "MODEL"
LABEL = "LABEL"

# Just used to simplify sorting.
LARGE_INTEGER = 0xffffffff


class AbstractContainer(Rollbackable) :
  """
  Base class for all objects that store abstract items with GET and from which
  such may be retrieved for PUT.  Meta data from the lens may be used to aid
  storage and retrieval of items in the container (e.g. labels, keys, concrete
  residue).

  We must be careful to allow the state of the container to be correctly
  captured and re-instated to facilitate efficient rollback.  In the worst
  case, we can deep-copy the state, but we should avoid this by considering
  how a particular container modifies its state during GET and PUT.  For
  example, with a list we may add or remove an item but not make finer changes
  to an item; with a general class, however, this may not be the case.
  """

  def __new__(cls, *args, **kargs) :
    self = super(AbstractContainer, cls).__new__(cls, *args, **kargs)
    # Initialise some vars regardless of __init__ being called.
    self._container_lens = None
    self._label = None
    self._alignment_mode = None

    # Return the pre-__init__ instance.
    return self

  def set_container_lens(self, lens) :
    # XXX: It's not the lens that's import, just that we wish to look at it to
    # set some defaults.
    # Called when put() is preparing the container for PUTting to associate it with its container lens.
    self._container_lens = lens
    # Default to MODEL alignment
    self._alignment_mode = self._container_lens.options.alignment or MODEL

  def set_label(self, label, overwrite=False) :
    if not overwrite :
      assert_msg(isinstance(label, str), "The container label must be a string --- at least for the time being.")
      if label and self._label:
        raise Exception("Container already has a label defined: %s" % self._label)
    self._label = label

  def get_label(self) :
    return self._label


  #
  # Can overload these for more control (e.g. tentative PUT/CREATE if non-deterministic)
  # 

  def get_and_store_item(self, lens, concrete_input_reader) :
    """Called by lenses that store items from sub-lenses in the container (e.g. And)."""
    # Note, here the lens may not have a type, though may still return an item
    # that was GOT from a sub-lens
    item = lens.get(concrete_input_reader, self)
    if has_value(item) :
      # Note, we check the actual item for is_label rather than the lens that
      # returned it, since the is_label lens may actually be a sublens.
      if item._meta_data.is_label :
        self.set_label(item) # Store item as label.
      else :
        self.store_item(item, lens, concrete_input_reader)


  def consume_and_put_item(self, lens, concrete_input_reader) :
    """Called by lenses that put items from the container into sub-lenses (e.g. And)."""
    assert(lens.has_type())
    assert_msg(has_value(self._container_lens), "Our container has not been associated with a container type lens.")
   
    # Handle is_label lens
    if lens.options.is_label:
      if not self._label :
        raise NoTokenToConsumeException("There was no item as this container's label to PUT.")
      output = lens.put(self._label, concrete_input_reader, None)
      self._label = None
      return output

    # Get candidates to PUT
    candidates = self.get_put_candidates(lens, concrete_input_reader)

    d("Unfiltered candidates: %s" % candidates)
    
    # Filter and sort them appropriately for our context (e.g. the lens, the
    # alignment mode and the current input postion.
    candidates = self.filter_and_sort_candidate_items(candidates, lens, concrete_input_reader)
    
    d("Filtered candidates: %s" % candidates)

    for candidate in candidates :
      try :
        # XXX : Overkill to copy initial state every time within automatic_rollback.
        with automatic_rollback(concrete_input_reader) :
          output = lens.put(candidate, concrete_input_reader, None)
          self.remove_item(lens, candidate)
          return output
      except LensException:
        pass

    # Didn't PUT any token.
    raise NoTokenToConsumeException()


  def filter_and_sort_candidate_items(self, candidate_items, lens, concrete_input_reader) :
    """
    In some cases we can whittle down the candidate list based on properties
    of the lens, the container's alignment mode, and our position in the
    concrete_input_reader.  We can then sort them, to give preference for which
    will be tried firstmost in the consume_and_put_item() function.
    """

    # Handle a static label lens, in which the candidate choice is
    # straightforward - here, for flexibility, we assume several items may share
    # a static label.
    if has_value(lens.options.label) :
      d("Using static label: '%s'" % lens.options.label)
      # XXX: Feels a bit of a hack to use attr_label, so will think more
      # generally about this.
      valid_candidates = [item for item in candidate_items if lens.options.label in [item._meta_data.label, item._meta_data.attr_label]]
      return valid_candidates

    # Handle MODEL alignment (i.e. PUT will be in order of items in the abstract
    # model).
    if self._alignment_mode == MODEL :
      # By definition, items are already in that order.
      if len(candidate_items) > 0:
        return [candidate_items[0]]
      else:
        return [] 
    
    # Handle SOURCE alignment.
    if self._alignment_mode == SOURCE :
      # Copy candidate_items.
      def get_key(item) :
        if has_value(item._meta_data.concrete_start_position) :
          return item._meta_data.concrete_start_position
        return LARGE_INTEGER # To ensure new items go on the end.
      
      # Sort the candiates by their source order - if they have meta on their
      # source position.
      sorted_candidate_items = sorted(candidate_items, key = get_key)
      return sorted_candidate_items
    # TODO: LABEL alignment mode
    
    raise Exception("Unknown alignment mode: %s" % self._alignment_mode)


  #
  # Must overload these.
  # 

  def get_put_candidates(self, lens, concrete_input_reader) :
    """Returns a list of candidate items that could be PUT into the lens."""
    raise NotImplementedError()
  
  def remove_item(self, lens, item) :
    raise NotImplementedError()

  def store_item(self, item, lens, concrete_input_reader) :
    raise NotImplementedError()

  def prepare_for_put(self) :
    """Called prior to PUT to ensure, for example, any internal containers are wrapped as AbstractContainers."""
    pass

  def unwrap(self) :
    """Unwrap to native python type where appropriate: e.g. for list and dict.""" 
    raise NotImplementedError()

  def is_fully_consumed(self) :
    """
    A container-specific sanity check to see if this container has been fully
    consumed (i.e. all items have been PUT).

    Note, since the label item may not always be consumed (only if an is_label
    lens is used), this should be discounted.
    """
    raise NotImplementedError()
    


class ListContainer(AbstractContainer) :
  """Most basic container, for storing items in a list."""

  def __new__(cls, *args, **kargs) :
    self = super(ListContainer, cls).__new__(cls, *args, **kargs)
    self.container_item = []
    return self
  
  def __init__(self, container_item) :
    assert isinstance(container_item, list)
    self.container_item = container_item
  
    assert_msg(isinstance(self.container_item, list))

    # Ensure our items can carry meta data (for algorithmic convenience) and be careful
    # to preserve the incoming lists meta data by modifying it in place.
    # Perhaps this can be done in AbstractContainer
    for index, item in enumerate(self.container_item) :
      self.container_item[index] = enable_meta_data(item)
    
      
  def get_put_candidates(self, lens, concrete_input_reader) :
    return self.container_item
  
  def remove_item(self, lens, item) :
    #d("Removing item %s" % item)
    self.container_item.remove(item)

  def store_item(self, item, lens, concrete_input_reader) :
    self.container_item.append(item)

  
  def unwrap(self):
    return self.container_item

  def _get_state(self, copy_state=True) :
    state = [copy_state and copy.copy(self.container_item) or self.container_item, self._label]
    return state

  def _set_state(self, state, copy_state=True) :
    self.container_item = copy_state and copy.copy(state[0]) or state[0]
    self._label = state[1]

  def __str__(self) :
    return str(self.container_item)
  __repr__ = __str__
  
  def is_fully_consumed(self) :
    return len(self.container_item) == 0


class DictContainer(ListContainer) :
  """Allows a list of items with labels to be accessed as a native python dict."""
  
  def __init__(self, container_item) :
    """
    We actually use all the functionality of the ListContainer, which we
    can later unwrap to a dict, since items may carry labels in their
    meta data.
    """

    # Create new dict if not passed one to wrap.
    assert isinstance(container_item, dict)
    self.container_item = container_item

    # Create a list to hold the items.
    items_as_list = []

    # XXX: Perhaps this should got in prepare_for_put.
    # Now add the items to the list, updating their labels from keys.
    assert isinstance(container_item, dict)
    for key, item in container_item.iteritems() :
      item = enable_meta_data(item)
      item._meta_data.label = key
      items_as_list.append(item)

    super(DictContainer, self).__init__(items_as_list)

  # TODO: Choose default alignment mode in set_container_lens().

  def store_item(self, item, *args, **kargs) :
    if not has_value(item._meta_data.label) :
      raise LensException("%s expected item %s to have a label." % (self, item))
    super(DictContainer, self).store_item(item, *args, **kargs)

  def unwrap(self):
    # First unwrap to a list.
    items_as_list = super(DictContainer, self).unwrap()

    # Then convert to a dictionary, using item labels as keys.
    items_as_dict = {}
    for item in items_as_list :
      items_as_dict[item._meta_data.label] = item
    return items_as_dict


#
# LensObject and related classes.
#

# Note that later we may manipulate attribute actions based on arbitrary
# properties set on these two classes, such as specific label-attribute
# mappings, etc.

class Attribute(Properties) : 
  """
  This class is used for declaring an attribute of a LensObject
  
  The static counter is used so we can infer the order of declaration of attributes.
  """
  
  counter = 0
  def __init__(self, *args, **kargs) :
    super(Attribute, self).__init__(*args, **kargs)
    self._counter = self.__class__.counter
    self.__class__.counter += 1
    
    
class Container(Attribute) :
  """Used for declaring a container attribute of a LensObject."""
  pass


class LensObject(AbstractContainer) :
  """
  A container that stores labelled items as object attributes, providing
  simple conversion of labels to and from python identifiers to store as
  attributes.  For example an item with the label "Last Name" would be stored
  as an object attribute as "last_name".

  To enhance the default behaviour of a LensObject, it is possible to constrain
  valid attributes and to declare sub-containers (e.g. lists, dict, etc.) which
  can be used to store specific items.
  """

  # Used to help with re-generating labels from object attribute names.  The
  # class seems a good a place as any to store this globally-useful data.
  __cached_labels = {}

  
  def __new__(cls, *args, **kargs) :
    """
    It is important that container objects can be created without arguments
    (a la serialisation), so here we initialise important internal variables before the
    __init__ is called.
    """
    self = super(LensObject, cls).__new__(cls, *args, **kargs)
  
    # If sub-containers have been specified on the class, instantiate them on the instance.
    self._create_containers_and_attributes()

    # Exclude attributes that should not be considered as container state.
    self._set_excluded_attributes()

    return self

  #
  # GET related.
  #

  def store_item(self, item, lens, concrete_input_reader) :
    
    # First see if the item is to be stored in one of our containers.
    sub_container = self._get_item_sub_container(lens, item)
    if sub_container :
      d("Storing %s in container %s" % (item, sub_container))
      return sub_container.store_item(item, lens, concrete_input_reader)

    if not has_value(item._meta_data.label) :
      raise LensException("%s expected item %s to have a label." % (self, item))
    # TODO: If constrained attributes, check within set.
    setattr(self, self.map_label_to_identifier(item._meta_data.label), item)

  
  def unwrap(self):
    """We are both the container and the native object."""
    # XXX: Note, somewhere before PUT we must reciprocate this.
    # Unwrap any sub containers.
    for name, container in self._containers.iteritems() :
      setattr(self, name, container.unwrap())

    return self
 
  #
  # PUT related.
  #
  
  def set_container_lens(self, lens) :
    # TODO: Shall we call on sub containers - perhaps not, since can set alignment from props
    super(LensObject, self).set_container_lens(lens)
    # For a general class container, SOURCE alignment will be a more common default.
    # Maybe LABEL mode would be more suitable when I implement it.
    self._alignment_mode = self._container_lens.options.alignment or SOURCE
 

  def get_put_candidates(self, lens, concrete_input_reader) :
    # First see if the item is to be put from one of our containers.
    sub_container = self._get_item_sub_container(lens)
    if sub_container :
      d("Using sub container %s" % sub_container)
      return sub_container.get_put_candidates(lens, concrete_input_reader)

    # Now try to find our own candidates.
    d("Looking for own canidates. %s" % self.__dict__)
    candidates = []

    # Append all of our data attributes that are not None.
    for attr_name in self._get_attribute_names() :
      # Skip if the item has not been set.
      if attr_name not in self.__dict__ :
        continue

      item = self.__dict__[attr_name]
      if has_value(item) :
        candidates.append(item)
    
    return candidates

 
  def remove_item(self, lens, item) :
    # First see if the item is to be put from one of our containers.
    sub_container = self._get_item_sub_container(lens, item)
    if sub_container :
      d("Removing %s from %s" % (item, sub_container))
      sub_container.remove_item(lens, item)
      return
    
    d("Preparing to remove %s" % item)
    for attr_name, value in self.__dict__.iteritems() :
      if value is item :
        del self.__dict__[attr_name]
        return

    raise Exception("Failed to remove item %s from %s."% (item, self))


  def prepare_for_put(self) :
    """Here we ensure any raw containers are wrapped as AbstractContainers - the opposite of unwrap()."""
    # Ensure all items that may be PUT may carry meta data.
    self._enable_attributes_meta()
    
    # Note, if there is no raw_container, __new__ would have ensure we still have an empty container in _containers
    for name, container in self._containers.iteritems() :
      # We do not use getattr, since it may return a class attribute by same name.
      raw_container = get_instance_attr(self, name, None)
      if not has_value(raw_container):
        continue # Don't destroy empty container already prepared in __new__
      raw_container = enable_meta_data(raw_container)
      if raw_container :
        self._containers[name] = ContainerFactory.wrap_container(raw_container)

  def is_fully_consumed(self) :
    # Check if our items are consumed.
    for attribute_name in self._get_attribute_names() :
      if attribute_name in self.__dict__ and has_value(self.__dict__[attribute_name]) :
        return False

    # Then, check if at least one of our containers is not fully consumed.
    for sub_container in self._containers.itervalues() :
      if not sub_container.is_fully_consumed() :
        return False

    return True

  #
  # Identifer mapping.
  #

  def map_label_to_identifier(self, label) :
    """
    Tries to convert a typical label to a python identifier.  You may wish to
    overload this if you require more specialised conversion.
    """
    identifier = self._map_label_to_identifier(label)

    # Check we end up with a valid python keyword.
    if not re.match(r"[a-zA-Z][a-zA-Z0-9_]*$", identifier) :
      raise Exception("Cannot express label '%s' (converted to '%s') as a python identifer to set as an object attribute - you will have to specialise this functionality for your purposes." % (label, identifier)) 
   
    # Cache this conversion on the class, since it may be useful to improve
    # CREATED labels.
    self.__class__.__cached_labels[identifier] = label

    return identifier
  
  def map_identifier_to_label(self, identifier) :
    if identifier in self.__class__.__cached_labels :
      return self.__class__.__cached_labels[identifier]

    # We assume that an underscore represents a space.
    return self._map_identifier_to_label(identifier)

  # He he: Really we should use a lens for these mappings, but perhaps it's
  # easy enough to do it like this.
  def _map_label_to_identifier(self, label) :
    """ This might be overridded to specialise this functionality for a specific.  """
    identifier = label.lower()
    # XXX: We could pre-compile this regex.
    identifier = re.sub(r"[ ]+", "_", identifier)
    return identifier
  
  def _map_identifier_to_label(self, identifier) :
    """ This might be overridded to specialise this functionality for a specific.  """
    return identifier.replace("_", " ")

  #
  # Other internal functions.
  #

  def _create_containers_and_attributes(self) :
    """Create any sub-containers, if declared."""
    self._containers = {}
    self._constrained_attributes = {}
    for key, value in self.__class__.__dict__.iteritems() :
      # Handle containers.
      if isinstance(value, Container) :
        container_properties = value
        assert_msg(has_value(container_properties.type), "You must declare a type for the container definition '%s'." % key)
        container = ContainerFactory.create_container(container_properties.type)
        assert_msg(has_value(container), "Could not create an appropriate container for '%s'." % key)
        self._containers[key] = container

      # Handle explicit attributes, which will constainer those accepted and define CREATE order.
      elif isinstance(value, Attribute) :
        self._constrained_attributes[key] = value


  def _get_item_sub_container(self, lens, item=None) :
    # Note, for lens matching in the get direction we use item meta (since may
    # be returned from higer lens with no type, such as Or)
    # In the put direction, we can use lens, since will only put a token into a
    # type lens.
    #if has_value(item) :
    #  d(item)
    #  assert(item._meta_data.lens)

    for name, container in self._containers.iteritems() :
      container_properties = self.__class__.__dict__[name]
      d("looking for item to match lens %s" % lens)
      if has_value(container_properties.store_items_from_lenses) and lens in container_properties.store_items_from_lenses :
        return container
      elif has_value(item) and has_value(container_properties.store_items_from_lenses) and has_value(item._meta_data.lens) and item._meta_data.lens in container_properties.store_items_from_lenses :
        return container
      elif has_value(container_properties.store_items_of_type) and type(item) in container_properties.store_items_of_type :
        return container
      elif lens.has_type() and lens.type in container_properties.store_items_of_type :
        return container

    return None


  def _set_excluded_attributes(self) :
    """
    In the case where the user does not constrain which attributes are treated
    as container state (i.e. to be GOT and PUT by the lens), here we try to
    exclude any attributes that are obviously not such state.  This
    automatically builds a list of attributes to exclude from our container's
    state.  This should be called on object initialisation, before any model
    attributes are asigned.

    Note that any attribute that starts with an underscore will be excluded, so
    this aims to exclude any other attributes, such as sub-container attributes.
    """
    self._excluded_attributes = self.__dict__.keys() + self._containers.keys()


  def _get_attribute_names(self) :
    """Returns the names of attributes that are being used to hold data items."""
    
    # If the useable attributes have been declared, use their names here, in declaration order.
    if self._constrained_attributes :
      # Sort the items according to their declaration order - captured by a static counter in Attribute
      items = sorted(self._constrained_attributes.iteritems(), key=lambda x: x[1]._counter)
      attributes = [item[0] for item in items]
      return attributes

    # Otherwise, use every object attribute that is not excluded.
    attributes = []
    for attr_name in self.__dict__.keys() :
      if attr_name not in self._excluded_attributes and not attr_name.startswith("_"):
        attributes.append(attr_name)
    
    return attributes

  def _enable_attributes_meta(self) :
    """Enables meta on attributes that may be used as container state."""
    for attr_name in self._get_attribute_names() :
      if attr_name not in self.__dict__ :
        continue

      item = enable_meta_data(self.__dict__[attr_name])
      self.__dict__[attr_name] = item
      # Ensure the label of the item is updated to match the current attribute
      # name.  If our label has changed, we need to regenerate a label.
      current_label = item._meta_data.label
      if not (has_value(current_label) and self.map_label_to_identifier(current_label) == attr_name) :
        item._meta_data.label = self.map_identifier_to_label(attr_name)
        # XXX: Feels like a hack for now, to get around issue of static labels
        # being changed incorrectly in the same way as a dynamic label
        item._meta_data.attr_label = attr_name

  def _get_state(self, copy_state=True) :
    # Get our state.
    state = [copy_state and copy.copy(self.__dict__) or self.__dict__]
    
    # Then append state of our containers.
    for sub_container in self._containers.itervalues() :
      state.append(sub_container._get_state(copy_state=copy_state))
    
    return state

  def _set_state(self, state, copy_state=True) :
    # XXX: Is there any conflict here with broad use of __dict__?
    # Set our state.
    self.__dict__ = copy_state and copy.copy(state[0]) or state[0]
    
    # Then set the state of our containers.
    i = 1
    for sub_container in self._containers.itervalues() :
      sub_container._set_state(state[i], copy_state=copy_state)
      i += 1


class ContainerFactory:
  """
  Used to create appropriate containers for particular types of lens.  For
  example, lens.type == dict -> container_class == DictContainer.
  """

  @staticmethod
  def get_container_class(incoming_type) :
    """
    Returns the container class associated with this type, if there is one.
    """
    if incoming_type == None:
      return None

    # If type is already an AbstractContainer, return it.
    if issubclass(incoming_type, AbstractContainer) :
      return incoming_type

    # Handle associations with native python containers.
    if issubclass(incoming_type, list) :
      return ListContainer
    elif issubclass(incoming_type, dict) :
      return DictContainer
    
    return None


  @staticmethod
  def create_container(container_type) :
    """
    Tries to create an container appropriate for this lens and returns None
    if there is no associated container class.
    """
    # See if the lens type has a container class.
    container_class = ContainerFactory.get_container_class(container_type)
    
    if container_class == None:
      return None
    
    # We do not call the constructor, which may require args.
    return container_class.__new__(container_class)

  @staticmethod
  def wrap_container(incoming_object) :
    """Wraps a container if possible."""
    if not has_value(incoming_object) :
      return None
    
    if issubclass(type(incoming_object), AbstractContainer) :
      container = incoming_object
    else :
      #d("Wrapping %s" % incoming_object)
      container_class = ContainerFactory.get_container_class(type(incoming_object))
      if not has_value(container_class) :
        return None
     
      # Pass the raw data item to wrap.
      container = container_class(incoming_object)
   
    # Allow the container to prepare to be PUT (e.g. wrap any sub containers).
    container.prepare_for_put()
    # Set the (consumable) label of the container based on the item's current
    # label
    container.set_label(incoming_object._meta_data.label, overwrite=True)
    return container
